#!/usr/bin/env python

from __future__ import print_function
import os
import sys
import time

# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 1

# sleep a little bit or we will never see the asm in the configuration file
# and the message received just before we go to the established loop will be printed twice
time.sleep(1)

print('announce operational rpcq afi ipv4 safi unicast sequence %d' % counter)
print('announce operational rpcp afi ipv4 safi unicast sequence %d counter 200' % counter)
time.sleep(1)

counter += 1

print('announce operational apcq afi ipv4 safi unicast sequence %d' % counter)
print('announce operational apcp afi ipv4 safi unicast sequence %d counter 150' % counter)
time.sleep(1)

counter += 1

print('announce operational lpcq afi ipv4 safi unicast sequence %d' % counter)
print('announce operational lpcp afi ipv4 safi unicast sequence %d counter 250' % counter)
time.sleep(1)

while True:
	try:
		time.sleep(1)
		if counter % 2:
			print('announce operational adm afi ipv4 safi unicast advisory "this is dynamic message #%d"' % counter)
			sys.stdout.flush()
		else:
			print('announce operational asm afi ipv4 safi unicast advisory "we SHOULD not send asm from the API"')
			sys.stdout.flush()

		counter += 1
	except KeyboardInterrupt:
		pass
	except IOError:
		break



# encoding: utf-8
"""
parse_process.py

Created by Thomas Mangin on 2015-06-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
import uuid

from exabgp.configuration.core import Section

from exabgp.configuration.process.parser import encoder
from exabgp.configuration.process.parser import run


class ParseProcess (Section):
	syntax = \
		'process name-of-process {\n' \
		'   run /path/to/command with its args;\n' \
		'   encoder text|json;\n' \
		'}' \

	known = {
		'encoder': encoder,
		'run':     run,
	}

	action = {
		'encoder': 'set-command',
		'run':     'set-command',
	}

	name = 'process'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)
		self.processes = {}
		self._processes = []
		self.named = ''

	def clear (self):
		self.processes = {}
		self._processes = []

	def pre (self):
		self.named = self.tokeniser.line[1]
		if self.named in self._processes:
			return self.error.set('a process section called "%s" already exists' % self.named)
		self._processes.append(self.named)
		return True

	def post (self):
		difference = set(self.known.keys()).difference(self.scope.get().keys())
		if difference:
			return self.error.set('unset process sections: %s' % ', '.join(difference))
		self.processes.update({self.named: self.scope.pop()})
		return True

	def add_api (self):
		if not os.environ.get('exabgp_cli_pipe',''):
			return
		name = 'api-internal-cli-%x' % uuid.uuid1().fields[0]
		api = {
			name: {
				'run': [sys.executable, os.path.join(os.environ.get('PWD',''),sys.argv[0])],
				'encoder': 'text'
			}
		}
		self._processes.append(name)
		self.processes.update(api)



#!/usr/bin/env python
# encoding: utf-8
"""
parsing.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

import os
import sys
import glob

from exabgp.configuration.configuration import Configuration
from exabgp.configuration.check import check_neighbor

from exabgp.configuration.setup import environment
environ = environment.setup('')
environ.log.enable = True
environ.log.all = False
environ.log.configuration = False
environ.log.parser = False


class TestControl (unittest.TestCase):
	def setUp (self):
		location = os.path.abspath(os.path.join(os.path.dirname(__file__),'..','conf','*.conf'))
		self.files = glob.glob(location)

	# These files contains invalid attribute we can not parse
	skip = 'attributes.conf'

	def test_all_configuration (self):
		for filename in self.files:
			if filename.endswith(self.skip):
				continue
			print('-'*80)
			print(filename)
			print('='*80)
			sys.stdout.flush()
			configuration = Configuration([filename,])
			configuration.reload()
			self.assertEqual(check_neighbor(configuration.neighbors),True)
			del configuration

if __name__ == '__main__':
	unittest.main()



# encoding: utf-8
"""
scope.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys
import pprint

# from copy import deepcopy
from exabgp.vendoring import six
from exabgp.protocol.ip import IP
from exabgp.configuration.core.error import Error


if sys.version_info > (3,):
	long = int


class Scope (Error):
	def __init__ (self):
		Error.__init__(self)
		self._location = []
		self._added = set()
		self._all = {
			'template': {},
		}
		self._routes = []
		self._current = self._all

	def __repr__ (self):
		return pprint.pformat(self.__dict__,indent=3)

	def clear (self):
		self._location = []
		self._added = set()
		self._all = {
			'template': {},
		}
		self._routes = []
		self._current = self._all

	# building route list

	def get_routes(self):
		return self._routes

	def pop_routes(self):
		routes = self._routes
		self._routes = []
		return routes

	def extend_routes (self, value):
		self._routes.extend(value)

	# building nlri

	def append_route(self, value):
		return self._routes.append(value)

	def get_route(self):
		return self._routes[-1]

	def pop_route(self):
		return self._routes.pop()

	# context

	def enter (self,location):
		self._location.append(location)

	def leave (self):
		if not len(self._location):
			return ''  # This is signaling an issue to the caller without raising
		return self._location.pop()

	def location (self):
		return '/'.join(self._location)

	# context

	def to_context (self,name=''):
		self._current = self._all
		for context in self._location:
			if context not in self._current:
				self._current[context] = {}
			self._current = self._current[context]
		if name:
			self._current = self._current.setdefault(name,{})

	def pop_context (self,name):
		returned = self._all.pop(name)

		for inherit in returned.get('inherit',[]):
			if inherit not in self._all['template'].get('neighbor',{}):
				self.throw('invalid template name referenced')
			self.transfer(self._all['template']['neighbor'][inherit],returned)

		return returned

	# key / value

	def set (self, name, value):
		self._current[name] = value

	def attribute_add (self, name, data):
		# .add_and_merge() and not .add() is required
		# flow spec to have multiple keywords adding to the extended-community
		self.get_route().attributes.add_and_merge(data)
		if name not in self._added:
			self._added.add(name)

	def nlri_assign (self, name, command, data):
		self.get_route().nlri.assign(command,data)

	def nlri_add (self, name, command, data):
		self.get_route().nlri.add(data)

	def nlri_nexthop (self, name, data):
		self.get_route().nlri.nexthop = data

	def append (self, name, data):
		self._current.setdefault(name,[]).append(data)

	def extend (self, name, data):
		self._current.setdefault(name,[]).extend(data)

	def merge (self, name, data):
		for key in data:
			value = data[key]
			if key not in self._current:
				self.set(key,value)
			elif isinstance(value, list):
				self._current[key].extend(value)
			elif isinstance(value, dict):
				self.transfer(value,self._current[key])
			else:
				self.set(key,value)

	def inherit (self, data):
		return self.transfer(data,self._current)

	def transfer (self, source, destination):
		for key,value in six.iteritems(source):
			if key not in destination:
				destination[key] = value
			elif isinstance(source[key], list):
				destination[key].extend(value)
			elif isinstance(source[key], dict):
				if key not in destination:
					destination[key] = source[key]
				else:
					self.transfer(source[key],destination[key])
			elif isinstance(source[key], int):
				destination[key] = value
			elif isinstance(source[key], long):
				destination[key] = value
			elif isinstance(source[key],IP):
				destination[key] = value
			else:
				self.throw('can not recursively copy this type of data %s' % type(source[key]))

	def get (self, name='', default=None):
		if name:
			return self._current.get(name,default)
		return self._current

	def pop (self, name='', default=None):
		if name == '':
			return dict((k,self._current.pop(k)) for k in list(self._current))
		return self._current.pop(name,default)

	def template (self, template, name):
		return self._all['template'].get(template,{}).get(name,{})



# encoding: utf-8
"""
fsm.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys


# https://en.wikipedia.org/wiki/Border_Gateway_Protocol#Finite-state_machines

# ======================================================== Finite State Machine
#

class FSM (object):
	class STATE (int):
		if sys.version_info[0] < 3:
			__slots__ = ['code']

		IDLE        = 0x01
		ACTIVE      = 0x02
		CONNECT     = 0x04
		OPENSENT    = 0x08
		OPENCONFIRM = 0x10
		ESTABLISHED = 0x20

		names = {
			IDLE:           'IDLE',
			ACTIVE:         'ACTIVE',
			CONNECT:        'CONNECT',
			OPENSENT:       'OPENSENT',
			OPENCONFIRM:    'OPENCONFIRM',
			ESTABLISHED:    'ESTABLISHED',
		}

		codes = dict((name,code) for (code,name) in names.items())

		valid = list(names)

		def __init__ (self, code):
			if code not in self.valid:
				raise RuntimeError('invalid FSM code %s' % code)
			int.__init__(code)

		def __repr__ (self):
			return self.names.get(self,'INVALID 0x%s' % hex(self))

		def __str__ (self):
			return repr(self)

	IDLE        = STATE(0x01)
	ACTIVE      = STATE(0x02)
	CONNECT     = STATE(0x04)
	OPENSENT    = STATE(0x08)
	OPENCONFIRM = STATE(0x10)
	ESTABLISHED = STATE(0x20)

	# to: from
	transition = {
		IDLE:        [IDLE, ACTIVE, CONNECT, OPENSENT, OPENCONFIRM, ESTABLISHED],
		ACTIVE:      [IDLE, ACTIVE, OPENSENT],
		CONNECT:     [IDLE, CONNECT, ACTIVE],
		OPENSENT:    [CONNECT],
		OPENCONFIRM: [OPENSENT, OPENCONFIRM],
		ESTABLISHED: [OPENCONFIRM, ESTABLISHED],
	}

	def __init__ (self, peer, state):
		self.peer = peer
		self.state = state

	def change (self, state):
		# if self.state not in self.transition.get(state,[]):
		# 	raise RuntimeError ('invalid state machine transition (from %s to %s)' % (str(self.state),str(state)))
		self.state = state
		if self.peer.neighbor.api['fsm']:
			self.peer.reactor.processes.fsm(self.peer.neighbor,self)
		return self

	def __eq__ (self, other):
		return self.state == other

	def __neq__ (self, other):
		return not self.__eq__(other)

	def __repr__ (self):
		return 'FSM state %s' % self.state

	def name (self):
		return self.STATE.names.get(self.state,'INVALID')



# encoding: utf-8
"""
mac.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""


from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i


# ========================================================================== MAC
#

class MAC (object):

	__slots__ = ['mac','_packed']

	def __init__ (self, mac=None,packed=None):
		self.mac = mac
		self._packed = packed if packed else concat_bytes_i(character(int(_,16)) for _ in mac.split(":"))

	def __eq__ (self, other):
		return self.mac == other.mac

	def __neq__ (self, other):
		return self.mac != other.mac

	def __lt__ (self, other):
		raise RuntimeError('comparing MAC for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing MAC for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing MAC for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing MAC for ordering does not make sense')

	def __str__ (self):
		return ':'.join('%02X' % ordinal(_) for _ in self._packed)

	def __repr__ (self):
		return self.__str__()

	def pack (self):
		return self._packed

	# Orange code was returning 10 !
	def __len__ (self):
		return 6

	# XXX: FIXME: improve for better performance ?
	def __hash__ (self):
		return hash(str(self))

	@classmethod
	def unpack (cls, data):
		return cls(':'.join('%02X' % ordinal(_) for _ in data[:6]),data[:6])

	def json (self, compact=None):
		return '"mac": "%s"' % str(self)



# encoding: utf-8
"""
change.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""


class Source (object):
	UNSET         = 0
	CONFIGURATION = 1
	API           = 2
	NETWORK       = 3


class Change (object):
	SOURCE = Source.UNSET

	__slots__ = ['nlri','attributes','__index']

	@staticmethod
	def family_prefix (family):
		return '%02x%02x' % family

	def __init__ (self, nlri, attributes):
		self.nlri = nlri
		self.attributes = attributes
		# prevent multiple allocation of the index when calling .index()
		# storing the value at __init__ time causes api-attributes.sequence to fail
		# XXX: the NLRI content is half missing !!
		self.__index = ''

	def index (self):
		if not self.__index:
			self.__index = '%02x%02x' % self.nlri.family() + self.nlri.index()
		return self.__index

	def __eq__ (self, other):
		return self.nlri == other.nlri and self.attributes == other.attributes

	def __ne__ (self, other):
		return self.nlri != other.nlri or self.attributes != other.attributes

	def __lt__ (self, other):
		raise RuntimeError('comparing Change for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing Change for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing Change for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing Change for ordering does not make sense')

	def extensive (self):
		# If you change this you must change as well extensive in Update
		return "%s%s" % (str(self.nlri),str(self.attributes))

	def __repr__ (self):
		return self.extensive()

	def feedback (self):
		if self.nlri is not None:
			return self.nlri.feedback(self.nlri.action)
		return 'no check implemented for the family %s %s' % self.nlri.family()


class ConfigurationChange (Change):
	SOURCE = Source.CONFIGURATION


class APIChange (Change):
	SOURCE = Source.API


class NetworkChange (Change):
	SOURCE = Source.NETWORK



# encoding: utf-8
"""

Support for RFC 8092

Copyright (c) 2016 Job Snijders <job@ntt.net>
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.bgp.message.update.attribute import Attribute

from struct import pack
from struct import unpack


class LargeCommunity (Attribute):
	MAX = 0xFFFFFFFFFFFFFFFFFFFFFFFF

	cache = {}
	caching = True

	__slots__ = ['large_community','_str']

	def __init__ (self, large_community):
		self.large_community = large_community
		self._str = "%d:%d:%d" % unpack('!LLL', self.large_community)

	def __eq__ (self, other):
		return self.large_community == other.large_community

	def __ne__ (self, other):
		return self.large_community != other.large_community

	def __lt__ (self, other):
		return self.large_community < other.large_community

	def __le__ (self, other):
		return self.large_community <= other.large_community

	def __gt__ (self, other):
		return self.large_community > other.large_community

	def __ge__ (self, other):
		return self.large_community >= other.large_community

	def json (self):
		return "[ %d, %d , %d ]" % unpack('!LLL', self.large_community)

	def pack (self, negotiated=None):
		return self.large_community

	def __repr__ (self):
		return self._str

	def __len__ (self):
		return 12

	@classmethod
	def unpack (cls, large_community, negotiated):
		return cls(large_community)

	@classmethod
	def cached (cls, large_community):
		if not cls.caching:
			return cls(large_community)
		if large_community in cls.cache:
			return cls.cache[large_community]
		instance = cls(large_community)
		cls.cache[large_community] = instance
		return instance



# encoding: utf-8
"""
announce/label.py

Created by Thomas Mangin on 2017-07-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import NoNextHop

from exabgp.rib.change import Change

from exabgp.bgp.message import OUT

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.update.nlri.inet import INET
from exabgp.bgp.message.update.nlri.cidr import CIDR
from exabgp.bgp.message.update.attribute import Attributes

from exabgp.configuration.announce import ParseAnnounce
from exabgp.configuration.announce.ip import AnnounceIP

from exabgp.configuration.static.parser import prefix
from exabgp.configuration.static.parser import path_information


class AnnouncePath (ParseAnnounce):
	# put next-hop first as it is a requirement atm
	definition = [
		'label <15 bits number>',
	] + AnnounceIP.definition

	syntax = \
		'<safi> <ip>/<netmask> { ' \
		'\n   ' + ' ;\n   '.join(definition) + '\n}'

	known = dict(AnnounceIP.known,**{
		'path-information':    path_information,
	})

	action = dict(AnnounceIP.action,**{
		'path-information':    'nlri-set',
	})

	assign = dict(AnnounceIP.assign,**{
		'path-information':    'path_info',
	})

	name = 'path'
	afi = None

	def __init__ (self, tokeniser, scope, error, logger):
		ParseAnnounce.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		return True

	def _check (self):
		if not self.check(self.scope.get(self.name),self.afi):
			return self.error.set(self.syntax)
		return True

	@staticmethod
	def check (change,afi):
		if change.nlri.nexthop is NoNextHop \
			and change.nlri.action == OUT.ANNOUNCE \
			and change.nlri.afi == afi \
			and change.nlri.safi in (SAFI.unicast,SAFI.multicast):
			return False
		return True


def ip_unicast (tokeniser,afi,safi):
	ipmask = prefix(tokeniser)

	nlri = INET(afi,safi,OUT.ANNOUNCE)
	nlri.cidr = CIDR(ipmask.pack(),ipmask.mask)

	change = Change(
		nlri,
		Attributes()
	)

	while True:
		command = tokeniser()

		if not command:
			break

		action = AnnouncePath.action.get(command,'')

		if action == 'attribute-add':
			change.attributes.add(AnnouncePath.known[command](tokeniser))
		elif action == 'nlri-set':
			change.nlri.assign(AnnouncePath.assign[command],AnnouncePath.known[command](tokeniser))
		elif action == 'nexthop-and-attribute':
			nexthop,attribute = AnnouncePath.known[command](tokeniser)
			change.nlri.nexthop = nexthop
			change.attributes.add(attribute)
		else:
			raise ValueError('route: unknown command "%s"' % command)

	return [change]


@ParseAnnounce.register('unicast','extend-name','ipv4')
def unicast_v4 (tokeniser):
	return ip_unicast(tokeniser,AFI.ipv4,SAFI.unicast)


@ParseAnnounce.register('unicast','extend-name','ipv6')
def unicast_v6 (tokeniser):
	return ip_unicast(tokeniser,AFI.ipv6,SAFI.unicast)



#!/usr/bin/env python

import sys

match = {
	0: ['neighbor 127.0.0.1 update start',] * 20,
	1: \
		['neighbor 127.0.0.1 announced route %d%d.0.0.0/8 next-hop 1.1.1.1 origin igp as-path [ 1 2 3 4 ] med 100' %
			(number,number) for number in range(1,10)
		] \
		+ \
		['neighbor 127.0.0.1 announced route eor %d/%d (%s)' % (afi,safi,text) for (afi,safi,text) in \
			(
				(1,1,'ipv4 unicast'),
				(1,2,'ipv4 multicast'),
				(1,4,'ipv4 nlri-mpls'),
				(1,128,'ipv4 mpls-vpn'),
				(1,133,'ipv4 flow'),
				(1,134,'ipv4 flow-vpn'),
				(2,1,'ipv6 unicast'),
				(2,128,'ipv6 mpls-vpn'),
				(2,133,'ipv6 flow'),
				(2,134,'ipv6 flow-vpn'),
				(25,65,'l2vpn vpls')
			)
		],
	2: ['neighbor 127.0.0.1 update end',] * 20,
	3: ['',] * 20,
}

count = 0

while True:
	line = sys.stdin.readline().strip()
	#print >> sys.stderr, '[%s]' % line.replace('\n','\\n'), start, end

	options = match[count % 4]

	if line in options:
		sys.stderr.write('%-3d ok      %s\n' % (count,line))
		sys.stderr.flush()
		count += 1
		continue

	sys.stderr.write('%-3d failed  %s\n' % (count,line))
	# sys.stderr.write("---" + line + "---\n")
	# sys.stderr.write("---" + str(options) + "---\n")
	sys.stderr.flush()
	count += 1



#!/usr/bin/env python
# encoding: utf-8
"""
control.py

Created by Thomas Mangin on 2015-01-01.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time
import socket
import unittest


from exabgp.configuration.setup import environment
environment.setup('')


def speak (name, data):
	time.sleep(0.005)
	try:
		sock = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
		sock.connect(name)
		sock.sendall(data)
	except socket.error:
		pass


class TestControl (unittest.TestCase):
	def setUp (self):
		pass

	def test_failed_creation (self):
		pass
		# control = Control()
		# try:
		# 	result = control.init()
		# 	self.assertFalse(result)
		# except IOError:
		# 	# could not write in the location
		# 	pass
		# finally:
		# 	control.cleanup()

	def validate (self, message, check):
		pass
		# name = tempfile.mktemp()
		# control = Control(name,False)
		# try:
		# 	result = control.init()
		# 	self.assertTrue(result)
		#
		# 	p = Process(target=speak, args=(name,message))
		# 	p.start()
		#
		# 	string = control.loop()
		# 	self.assertEqual(string, check)
		# 	p.join()
		# finally:
		# 	control.cleanup()
		# 	del control

	def test_no_newline (self):
		self.validate('x','')

	def test_one_newline (self):
		self.validate('x\n','x')

	def test_two_newline (self):
		self.validate('-\nx\n','x')

	def test_leftover (self):
		self.validate('-\nx\n-','x')

if __name__ == '__main__':
	unittest.main()



# encoding: utf-8
"""
ip/__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket

from exabgp.util import ordinal

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.protocol.ip.netmask import NetMask

# XXX: The IP,Range and CIDR class API are totally broken, fix it.
# XXX: many of the NLRI classes constructor also need correct @classmethods


# =========================================================================== IP
#

class IPSelf (object):
	def __init__ (self, afi):
		self.afi = afi

	def __repr__ (self):
		return 'self'

	def top (self, negotiated, afi=AFI.undefined):
		return negotiated.nexthopself(afi).top()

	def ton (self,negotiated, afi=AFI.undefined):
		return negotiated.nexthopself(afi).ton()

	def pack (self, negotiated):
		return negotiated.nexthopself(self.afi).ton()

	def index (self):
		return 'self-' + AFI.names[self.afi]


class IP (object):
	afi = None  # here for the API, changed in init which does not change this
	_known = dict()

	_UNICAST = SAFI.unicast
	_MULTICAST = SAFI.multicast

	_multicast_range = set(range(224,240))  # 239

	# deprecate the string API in favor of top()

	def __init__ (self):
		raise RuntimeError("You should use IP.create() to use IP")

	def init (self, string, packed=None):
		# XXX: the str should not be needed
		self._string = string
		self._packed = IP.pton(string) if packed is None else packed
		self.afi = IP.toafi(string)
		return self

	def __iter__ (self):
		for letter in self._string:
			yield letter

	@staticmethod
	def pton (ip):
		return socket.inet_pton(IP.toaf(ip),ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET if len(data) == 4 else socket.AF_INET6,data)

	def top (self, negotiated=None, afi=AFI.undefined):
		return self._string

	@staticmethod
	def toaf (ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			return socket.AF_INET6
		if '.' in ip:
			return socket.AF_INET
		raise ValueError('unrecognised ip address %s' % ip)

	@staticmethod
	def toafi (ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			return AFI.ipv6
		if '.' in ip:
			return AFI.ipv4
		raise ValueError('unrecognised ip address %s' % ip)

	@staticmethod
	def tosafi (ip):
		if ':' in ip:
			# XXX: FIXME: I assume that ::FFFF:<ip> must be treated unicast
			# if int(ip.split(':')[-1].split('.')[0]) in IP._multicast_range:
			return SAFI.unicast
		elif '.' in ip:
			if int(ip.split('.')[0]) in IP._multicast_range:
				return SAFI.multicast
			return SAFI.unicast
		raise ValueError('unrecognised ip address %s' % ip)

	def ipv4 (self):
		return True if len(self._packed) == 4 else False

	def ipv6 (self):
		return False if len(self._packed) == 4 else True

	def address (self):
		value = 0
		for char in self._packed:
			value <<= 8
			value += ordinal(char)
		return value

	@staticmethod
	def length (afi):
		return 4 if afi == AFI.ipv4 else 16

	def index (self):
		return self._packed

	def pack (self):
		return self._packed

	def ton (self, negotiated=None, afi=AFI.undefined):
		return self._packed

	def __repr__ (self):
		return self._string

	def __eq__ (self, other):
		if not isinstance(other, IP):
			return False
		return self._packed == other._packed

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __lt__ (self, other):
		return self._packed < other._packed

	def __le__ (self, other):
		return self._packed <= other._packed

	def __gt__ (self, other):
		return self._packed > other._packed

	def __ge__ (self, other):
		return self._packed >= other._packed

	def __hash__ (self):
		return hash((self.__class__.__name__, self._packed))

	@classmethod
	def klass (cls, ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			afi = IPv6.afi
		elif '.' in ip:
			afi = IPv4.afi
		else:
			raise ValueError('can not decode this ip address : %s' % ip)
		if afi in cls._known:
			return cls._known[afi]

	@classmethod
	def create (cls, string, packed=None, klass=None):
		if klass:
			return klass(string,packed)
		return cls.klass(string)(string,packed)

	@classmethod
	def register (cls):
		cls._known[cls.afi] = cls

	@classmethod
	def unpack (cls, data, klass=None):
		return cls.create(IP.ntop(data),data,klass)


# ======================================================================== Range
#

class IPRange (IP):
	def __init__ (self, ip, mask):
		IP.init(self,ip)
		self.mask = NetMask.create(mask,IP.toafi(ip))

	@classmethod
	def create (klass, ip, mask):
		return klass(ip,mask)

	def __repr__ (self):
		if (self.ipv4() and self.mask == 32) or \
		   (self.ipv6() and self.mask == 128):
			return super(IPRange, self).__repr__()
		else:
			return '%s/%d' % (self.top(), int(self.mask))


# ==================================================================== NoNextHop
#

class _NoNextHop (object):
	packed = ''

	def pack (self, data, negotiated=None):
		return ''

	def index (self):
		return ''

	def ton (self, negotiated=None, afi=AFI.undefined):
		return ''

	def __str__ (self):
		return 'no-nexthop'

	def __deepcopy__(self, _):
		return self

	def __copy__(self, _):
		return self


NoNextHop = _NoNextHop()


# ========================================================================= IPv4
#

class IPv4 (IP):
	# lower case to match the class Address API
	afi = AFI.ipv4

	__slots__ = []

	def __init__ (self, string, packed=None):
		self.init(string,packed if packed else IP.pton(string))

	def __len__ (self):
		return 4

	def unicast (self):
		return not self.multicast()

	def multicast (self):
		return ordinal(self._packed[0]) in set(range(224,240))  # 239 is last

	def ipv4 (self):
		return True

	def ipv6 (self):
		return False

	@staticmethod
	def pton (ip):
		return socket.inet_pton(socket.AF_INET,ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET,data)

	# klass is a trick for subclasses of IP/IPv4 such as NextHop / OriginatorID
	@classmethod
	def unpack (cls, data, klass=None):
		ip = socket.inet_ntop(socket.AF_INET,data)
		if klass:
			return klass(ip,data)
		return cls(ip,data)


IPv4.register()


# ========================================================================= IPv6
#

class IPv6 (IP):
	# lower case to match the class Address API
	afi = AFI.ipv6

	__slots__ = []

	def __init__ (self, string, packed=None):
		self.init(string,packed if packed else socket.inet_pton(socket.AF_INET6,string))

	def __len__ (self):
		return 16

	def ipv4 (self):
		return False

	def ipv6 (self):
		return True

	def unicast (self):
		return True

	def multicast (self):
		return False

	@staticmethod
	def pton (ip):
		return socket.inet_pton(socket.AF_INET6,ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET6,data)

	@classmethod
	def unpack (cls, data, klass=None):
		ip6 = socket.inet_ntop(socket.AF_INET6,data)
		if klass:
			return klass(ip6)
		return cls(ip6)


IPv6.register()



# encoding: utf-8
"""
delay.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2017-2017 Exa Networks. All rights reserved.
"""

import time


# ======================================================================== Delay
# Exponential backup for outgoing connection

class Delay (object):
	def __init__ (self):
		self._time = time.time()
		self._next = 0

	def reset (self):
		self._time = time.time()
		self._next = 0

	def increase (self):
		self._time = time.time() + self._next
		self._next = min(int(1 + self._next * 1.2),60)

	def backoff (self):
		return self._time > time.time()



# encoding: utf-8
"""
format.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import coroutine


def formated (line):
	changed_line = '#'
	new_line = line.strip().replace('\t',' ').replace(']',' ]').replace('[','[ ').replace(')',' )').replace('(','( ').replace(',',' , ')
	while new_line != changed_line:
		changed_line = new_line
		new_line = new_line.replace('  ',' ')
	return new_line


# convert special caracters

@coroutine.join
def unescape (string):
	start = 0
	while start < len(string):
		pos = string.find('\\', start)
		if pos == -1:
			yield string[start:]
			break
		yield string[start:pos]
		pos += 1
		esc = string[pos]
		if esc == 'b':
			yield '\b'
		elif esc == 'f':
			yield '\f'
		elif esc == 'n':
			yield '\n'
		elif esc == 'r':
			yield '\r'
		elif esc == 't':
			yield '\t'
		elif esc == 'u':
			yield character(int(string[pos + 1:pos + 5], 16))
			pos += 4
		else:
			yield esc
		start = pos + 1


# A coroutine which return the producer token, or string if quoted from the stream

def tokens (stream):
	spaces = [' ','\t','\r','\n']
	strings = ['"', "'"]
	syntax = [',','[',']']
	eol = [';','{','}']
	comment = ['#',]

	nb_lines = 0

	for letters in stream:
		line = unescape(letters)
		parsed = []
		nb_chars = 0
		nb_lines += 1
		quoted = ''
		word = ''
		for char in line:
			if char in comment:
				if quoted:
					word += char
					nb_chars += 1
				else:
					if word:
						parsed.append((nb_lines,nb_chars,char))
						word = ''
					break

			elif char in eol:
				if quoted:
					word += char
					nb_chars += 1
				else:
					if word:
						parsed.append((nb_lines,nb_chars-len(word),word))
						word = ''
					parsed.append((nb_lines,nb_chars,char))
					nb_chars += 1
					yield parsed
					parsed = []

			elif char in syntax:
				if quoted:
					word += char
				else:
					if word:
						parsed.append((nb_lines,nb_chars-len(word),word))
						word = ''
					parsed.append((nb_lines,nb_chars,char))
				nb_chars += 1

			elif char in spaces:
				if quoted:
					word += char
				elif word:
					parsed.append((nb_lines,nb_chars-len(word),word))
					word = ''
				nb_chars += 1

			elif char in strings:
				# word += char
				if quoted == char:
					quoted = ''
					parsed.append((nb_lines,nb_chars-len(word),word))
					word = ''
				else:
					quoted = char
				nb_chars += 1

			else:
				word += char
				nb_chars += 1

		if parsed:
			raise ValueError('invalid syntax')



#!/usr/bin/env python

from __future__ import print_function
import sys
import time
import random

def write (data):
	sys.stdout.write(data + '\n')
	sys.stdout.flush()

def main ():
	if len(sys.argv) < 2:
		print("%s <number of routes> <updates per second thereafter>")
		sys.exit(1)

	initial = sys.argv[1]
	thereafter = sys.argv[2]

	if not initial.isdigit() or not thereafter.isdigit():
		write('please give valid numbers')
		sys.exit(1)

	# Limit to sane numbers :-)
	number = int(initial) & 0x00FFFFFF
	after = int(thereafter) & 0x0000FFFF

	range1 = (number >> 16) & 0xFF
	range2 = (number >>  8) & 0xFF
	range3 = (number      ) & 0xFF

	ip = {}
	nexthop = ['%d.%d.%d.%d' % (random.randint(1,200),random.randint(0,255),random.randint(0,255),random.randint(0,255)) for _ in range(200)]

	for ip1 in range(0,range1):
		for ip2 in range(0,256):
			for ip3 in range(0,256):
				generated = '%d.%d.%d.%d' % (random.randint(1,200),ip1,ip2,ip3)
				ip[generated] = random.choice(nexthop)

	for ip2 in range (0,range2):
		for ip3 in range (0,256):
			generated = '%d.%d.%d.%d' % (random.randint(1,200),range1,ip2,ip3)
			ip[generated] = random.choice(nexthop)

	for ip3 in range (0,range3):
		generated = '%d.%d.%d.%d' % (random.randint(1,200),range1,range2,ip3)
		ip[generated] = random.choice(nexthop)

	count = 0

	# initial table dump
	for k,v in ip.iteritems():
		count += 1
		write('announce route %s next-hop %s med 1%02d as-path [ 100 101 102 103 104 105 106 107 108 109 110 ]' % (k,v,len(k)))
		if count % 10000 == 0:
			sys.stderr.write('initial : announced %d\n' % count)

	count &= 0xFFFFFFFe

	# modify routes forever
	while True:
		now = time.time()
		changed = {}

		for k,v in ip.iteritems():
			changed[k] = v
			if not random.randint(0,after):
				break

		for k,v in changed.iteritems():
			count += 2
			write('withdraw route %s next-hop %s med 1%02d as-path [ 100 101 102 103 104 105 106 107 108 109 110 ]' % (k,v,len(k)))
			ip[k] = random.choice(nexthop)
			write('announce route %s next-hop %s med 1%02d as-path [ 100 101 102 103 104 105 106 107 108 109 110 ]' % (k,ip[k],len(k)))
			if count % 100 == 0:
				sys.stderr.write('updates : announced %d\n' % count)


		time.sleep(time.time()-now+1.0)

if __name__ == '__main__':
	try:
		main()
	except KeyboardInterrupt:
		pass



# encoding: utf-8
"""
location.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# ===================================================================== Location
# file location


class Location (object):
	def __init__ (self, index_line=0, index_column=0, line=''):
		self.line = line
		self.index_line = index_line
		self.index_column = index_column

	def clear (self):
		self.index_line = 0
		self.index_column = 0
		self.line = ''


class Error (Exception):
	tabsize = 3
	syntax = ''

	def __init__ (self, location, message, syntax=''):
		self.line = location.line.replace('\t',' '*self.tabsize)
		self.index_line = location.index_line
		self.index_column = location.index_column + (self.tabsize-1) * location.line[:location.index_column].count('\t')

		self.message = '\n\n'.join((
			'problem parsing configuration file line %d position %d' % (location.index_line,location.index_column+1),
			'error message: %s' % message.replace('\t',' '*self.tabsize),
			'%s%s' % (self.line,'-' * self.index_column + '^')
		))
		# allow to give the right syntax in using Raised
		if syntax:
			self.message += '\n\n' + syntax

		Exception.__init__(self)

	def __repr__ (self):
		return self.message



from exabgp.vendoring import six

from exabgp.protocol.family import AFI
from .connection import Connection
from .tcp import create,bind
from .tcp import connect
from .tcp import MD5
from .tcp import nagle
from .tcp import TTL
from .tcp import TTLv6
from .tcp import asynchronous
from .tcp import ready
from .error import NetworkError


class Outgoing (Connection):
	direction = 'outgoing'

	def __init__ (self, afi, peer, local, port=179,md5='',md5_base64=False, ttl=None):
		Connection.__init__(self,afi,peer,local)

		self.logger.debug('attempting connection to %s:%d' % (self.peer,port),self.session())

		self.ttl = ttl
		self.afi = afi
		self.md5 = md5
		self.port = port

		try:
			self.io = create(afi)
			MD5(self.io,self.peer,port,md5,md5_base64)
			if afi == AFI.ipv4:
				TTL(self.io, self.peer, self.ttl)
			elif afi == AFI.ipv6:
				TTLv6(self.io, self.peer, self.ttl)
			if local:
				bind(self.io,self.local,afi)
			asynchronous(self.io, self.peer)
			connect(self.io,self.peer,port,afi,md5)
			if not self.local:
				self.local = self.io.getsockname()[0]
			self.success()
			self.init = True
		except NetworkError as exc:
			self.init = False
			self.close()
			self.logger.debug('connection to %s:%d failed, %s' % (self.peer,port,str(exc)),self.session())

	def establish (self):
		if not self.init:
			yield False
			return

		generator = ready(self.io)
		for connected in generator:
			if not connected:
				yield False
				continue
			yield True
		yield False
		# self.io MUST NOT be closed here, it is closed by the caller
		return


		nagle(self.io,self.peer)
		# Not working after connect() at least on FreeBSD TTL(self.io,self.peer,self.ttl)
		yield True



# encoding: utf-8
"""
tcpflags.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.resource import BitResource


# ====================================================================== TCPFlag
# https://www.iana.org/assignments/tcp-header-flags

class TCPFlag (BitResource):
	NAME = 'tcp flag'

	FIN    = 0x01
	SYN    = 0x02
	RST    = 0x04
	PUSH   = 0x08
	ACK    = 0x10
	URGENT = 0x20

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'FIN':    FIN,
		'SYN':    SYN,
		'RST':    RST,
		'PUSH':   PUSH,
		'ACK':    ACK,
		'URGENT': URGENT,
	}.items())

	names = dict([(r,l) for (l,r) in codes.items()])



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i


# ===================================================================== PathInfo
# RFC draft-ietf-idr-add-paths-09

class PathInfo (object):

	__slots__ = ['path_info']

	def __init__ (self, packed=None, integer=None, ip=None):
		if packed:
			self.path_info = packed
		elif ip:
			self.path_info = concat_bytes_i(character(int(_)) for _ in ip.split('.'))
		elif integer:
			self.path_info = concat_bytes_i(character((integer >> offset) & 0xff) for offset in [24,16,8,0])
		else:
			self.path_info = b''
		# sum(int(a)<<offset for (a,offset) in zip(ip.split('.'), range(24, -8, -8)))

	def __eq__ (self, other):
		return self.path_info == other.path_info

	def __neq__ (self, other):
		return self.path_info != other.path_info

	def __lt__ (self, other):
		raise RuntimeError('comparing PathInfo for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing PathInfo for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing PathInfo for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing PathInfo for ordering does not make sense')

	def __len__ (self):
		return len(self.path_info)

	def json (self):
		if self.path_info:
			return '"path-information": "%s"' % '.'.join([str(ordinal(_)) for _ in self.path_info])
		return ''

	def __repr__ (self):
		if self.path_info:
			return ' path-information %s' % '.'.join([str(ordinal(_)) for _ in self.path_info])
		return ''

	def pack (self):
		if self.path_info:
			return self.path_info
		return b'\x00\x00\x00\x00'


PathInfo.NOPATH = PathInfo()



from exabgp.util.errstr import errstr

from .connection import Connection
from .tcp import nagle
from .tcp import asynchronous
from .error import NetworkError
from .error import NotConnected

from exabgp.bgp.message import Notify


class Incoming (Connection):
	direction = 'incoming'

	def __init__ (self, afi, peer, local, io):
		Connection.__init__(self,afi,peer,local)

		self.logger.debug('connection from %s' % self.peer,'network')

		try:
			self.io = io
			asynchronous(self.io, self.peer)
			nagle(self.io,self.peer)
			self.success()
		except NetworkError as exc:
			self.close()
			raise NotConnected(errstr(exc))

	def notification (self, code, subcode, message):
		try:
			notification = Notify(code,subcode,message).message()
			for boolean in self.writer(notification):
				yield False
			self.close()
		except NetworkError:
			pass  # This is only be used when closing session due to unconfigured peers - so issues do not matter



# encoding: utf-8
"""
sr/ipv6sid.py

Created by Evelio Vila 2017-02-16
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""
from struct import pack

from exabgp.util import concat_bytes
from exabgp.protocol.ip import IP

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.sr.prefixsid import PrefixSid

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |       Type    |             Length            |   RESERVED    |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |            RESERVED           |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
# |                                                               |
# |                        IPv6 SID (16 octets)                   |
# |                                                               |
# |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 3.2.  IPv6 SID


@PrefixSid.register()
class SrV6Sid(object):
	TLV = 2
	LENGTH = 19

	def __init__ (self,v6sid,packed=None):
		self.v6sid = v6sid
		self.packed = self.pack()

	def __repr__ (self):
		return "sr-v6-sid %s" % (self.v6sid)

	def pack (self):
		return concat_bytes(
			pack('!B', self.TLV),
			pack('!H', self.LENGTH),
			pack('!B', 0),
			pack('!H', 0),
			IP.pton(self.v6sid)
		)

	@classmethod
	def unpack (cls,data,length):
		v6sid = -1
		if cls.LENGTH != length:
			raise Notify(3,5, "Invalid TLV size. Should be {0} but {1} received".format(cls.LENGTH, length))
		# RESERVED: 24 bit field for future use.  MUST be clear on
		# transmission an MUST be ignored at reception.
		data = data[3:19]
		v6sid = IP.unpack(data)
		return cls(v6sid=v6sid,packed=data)

	def json (self, compact=None):
		return '"sr-v6-sid": "%s"' % (self.v6sid)



#!/usr/bin/env python

import sys

count = 0

while True:
	line = sys.stdin.readline()
	if ' route' in line:
		count += 1
		if count % 100 == 0:
			sys.stderr.write('received %-10d\n' % count)
			sys.stderr.flush()



#!/usr/bin/env python

import os
import sys
import time
import syslog

def _prefixed (level, message):
	now = time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime())
	return "%s %-8s %-6d %s" % (now,level,os.getpid(),message)

syslog.openlog("ExaBGP")

# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 0

while True:
	try:
		line = sys.stdin.readline().strip()
		if line == "":
			counter += 1
			if counter > 100:
				break
			continue

		counter = 0

		syslog.syslog(syslog.LOG_ALERT, _prefixed('INFO',line))
	except KeyboardInterrupt:
		pass
	except IOError:
		# most likely a signal during readline
		pass



# encoding: utf-8
"""
line/watchdog.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.reactor.api.command.command import Command
from exabgp.reactor.api.command.limit import match_neighbors
from exabgp.reactor.api.command.limit import extract_neighbors

from exabgp.protocol.ip import NoNextHop
from exabgp.bgp.message import OUT
from exabgp.bgp.message.update.attribute import NextHop

from exabgp.configuration.static import ParseStaticRoute


def register_announce ():
	pass


@Command.register('text','announce route')
def announce_route (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_route(command)
			if not changes:
				self.log_failure('command could not parse route in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				if not ParseStaticRoute.check(change):
					self.log_message('invalid route for %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					continue
				change.nlri.action = OUT.ANNOUNCE
				reactor.configuration.inject_change(peers,change)
				self.log_message('route added to %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','withdraw route')
def withdraw_route (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_route(command)
			if not changes:
				self.log_failure('command could not parse route in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				# Change the action to withdraw before checking the route
				change.nlri.action = OUT.WITHDRAW
				# NextHop is a mandatory field (but we do not require in)
				if change.nlri.nexthop is NoNextHop:
					change.nlri.nexthop = NextHop('0.0.0.0')

				if not ParseStaticRoute.check(change):
					self.log_message('invalid route for %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					continue
				if reactor.configuration.inject_change(peers,change):
					self.log_message('route removed from %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False
				else:
					self.log_failure('route not found on %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','announce vpls')
def announce_vpls (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_vpls(command)
			if not changes:
				self.log_failure('command could not parse vpls in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.ANNOUNCE
				reactor.configuration.inject_change(peers,change)
				self.log_message('vpls added to %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the vpls')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the vpls')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','withdraw vpls')
def withdraw_vpls (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_vpls(command)

			if not changes:
				self.log_failure('command could not parse vpls in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.WITHDRAW
				if reactor.configuration.inject_change(peers,change):
					self.log_message('vpls removed from %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False
				else:
					self.log_failure('vpls not found on %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the vpls')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the vpls')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','announce attribute')
@Command.register('text','announce attributes')
def announce_attributes (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_attributes(command,peers)
			if not changes:
				self.log_failure('command could not parse route in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.ANNOUNCE
				reactor.configuration.inject_change(peers,change)
				self.log_message('route added to %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','withdraw attributes')
def withdraw_attribute (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_attributes(command,peers)
			if not changes:
				self.log_failure('command could not parse route in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.WITHDRAW
				if reactor.configuration.inject_change(peers,change):
					self.log_message('route removed from %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False
				else:
					self.log_failure('route not found on %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
					yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the route')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','announce flow')
def announce_flow (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_flow(command)
			if not changes:
				self.log_failure('command could not parse flow in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.ANNOUNCE
				reactor.configuration.inject_change(peers,change)
				self.log_message('flow added to %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the flow')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the flow')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','withdraw flow')
def withdraw_flow (self, reactor, service, line):
	def callback ():
		try:
			descriptions,command = extract_neighbors(line)
			peers = match_neighbors(reactor.peers,descriptions)
			if not peers:
				self.log_failure('no neighbor matching the command : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			changes = self.api_flow(command)

			if not changes:
				self.log_failure('command could not parse flow in : %s' % command)
				reactor.processes.answer(service,'error')
				yield True
				return

			for change in changes:
				change.nlri.action = OUT.WITHDRAW
				if reactor.configuration.inject_change(peers,change):
					self.log_message('flow removed from %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				else:
					self.log_failure('flow not found on %s : %s' % (', '.join(peers) if peers else 'all peers',change.extensive()))
				yield False

			reactor.processes.answer_done(service)
		except ValueError:
			self.log_failure('issue parsing the flow')
			reactor.processes.answer(service,'error')
			yield True
		except IndexError:
			self.log_failure('issue parsing the flow')
			reactor.processes.answer(service,'error')
			yield True

	reactor.asynchronous.schedule(service,line,callback())
	return True


@Command.register('text','announce eor')
def announce_eor (self, reactor, service, command):
	def callback (self, command, peers):
		family = self.api_eor(command)
		if not family:
			self.log_failure("Command could not parse eor : %s" % command)
			reactor.processes.answer(service,'error')
			yield True
			return

		reactor.configuration.inject_eor(peers,family)
		self.log_message("Sent to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',family.extensive()))
		yield False

		reactor.processes.answer_done(service)

	try:
		descriptions,command = extract_neighbors(command)
		peers = match_neighbors(reactor.peers,descriptions)
		if not peers:
			self.log_failure('no neighbor matching the command : %s' % command)
			reactor.processes.answer(service,'error')
			return False
		reactor.asynchronous.schedule(service,command,callback(self,command,peers))
		return True
	except ValueError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False
	except IndexError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False


@Command.register('text','announce route-refresh')
def announce_refresh (self, reactor, service, command):
	def callback (self, command, peers):
		refresh = self.api_refresh(command)
		if not refresh:
			self.log_failure("Command could not parse route-refresh command : %s" % command)
			reactor.processes.answer(service,'error')
			yield True
			return

		reactor.configuration.inject_refresh(peers,refresh)
		self.log_message("Sent to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',refresh.extensive()))

		yield False
		reactor.processes.answer_done(service)

	try:
		descriptions,command = extract_neighbors(command)
		peers = match_neighbors(reactor.peers,descriptions)
		if not peers:
			self.log_failure('no neighbor matching the command : %s' % command)
			reactor.processes.answer(service,'error')
			return False
		reactor.asynchronous.schedule(service,command,callback(self,command,peers))
		return True
	except ValueError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False
	except IndexError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False


@Command.register('text','announce operational')
def announce_operational (self, reactor, service, command):
	def callback (self, command, peers):
		operational = self.api_operational(command)
		if not operational:
			self.log_failure("Command could not parse operational command : %s" % command)
			reactor.processes.answer(service,'error')
			yield True
			return

		reactor.configuration.inject_operational(peers,operational)
		self.log_message("operational message sent to %s : %s" % (
			', '.join(peers if peers else []) if peers is not None else 'all peers',operational.extensive()
			)
		)
		yield False
		reactor.processes.answer_done(service)

	if (command.split() + ['be','safe'])[2].lower() not in ('asm','adm','rpcq','rpcp','apcq','apcp','lpcq','lpcp'):
		reactor.processes.answer_done(service)
		return False

	try:
		descriptions,command = extract_neighbors(command)
		peers = match_neighbors(reactor.peers,descriptions)
		if not peers:
			self.log_failure('no neighbor matching the command : %s' % command)
			reactor.processes.answer(service,'error')
			return False
		reactor.asynchronous.schedule(service,command,callback(self,command,peers))
		return True
	except ValueError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False
	except IndexError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service,'error')
		return False



from exabgp.bgp.message.update.nlri.bgpls.tlvs.ipreach import IpReach
from exabgp.bgp.message.update.attribute.bgpls.prefix.igptags import IgpTags
from exabgp.bgp.message.update.attribute.bgpls.prefix.prefixmetric import PrefixMetric
from exabgp.bgp.message.update.nlri.bgpls.tlvs.ospfroute import OspfRoute
from exabgp.bgp.message.update.nlri.bgpls.tlvs.node import NodeDescriptor

import unittest


class TestTlvs(unittest.TestCase):

    def test_ip_reach(self,):
        data = b'\n\n\x00'
        tlv = IpReach.unpack(data)
        self.assertEqual(tlv.json(), '"ip-reachability-tlv": "10.0.0.0"')

    def test_igp_tags(self,):
        data = b'\x00\x00\xff\xfe'
        tlv = IgpTags.unpack(data, len(data))
        self.assertEqual(tlv.json(), '"igp-route-tags": [65534]')

    def test_prefix_metric(self,):
        data = b'\x00\x00\x00\x14'
        tlv = PrefixMetric.unpack(data, len(data))
        self.assertEqual(tlv.json(), '"prefix-metric": 20')

    def test_ospf_route_type(self,):
        data = b'\x04'
        tlv = OspfRoute.unpack(data)
        self.assertEqual(tlv.json(), '"ospf-route-type": 4')


class TestDescriptors(unittest.TestCase):

    def test_node_descriptor(self,):
        data = b'\x02\x00\x00\x04\x00\x00\xff\xfd\x02\x01\x00\x04\x00\x00\x00\x00\x02\x03\x00\x04\nq?\xf0'
        igp_type = 3
        descriptor, remain = NodeDescriptor.unpack(data, igp_type)
        self.assertEqual(descriptor.json(), '"autonomous-system": 65533')
        descriptor, remain = NodeDescriptor.unpack(remain, igp_type)
        self.assertEqual(descriptor.json(), '"bgp-ls-identifier": "0"')
        descriptor, remain = NodeDescriptor.unpack(remain, igp_type)
        self.assertEqual(descriptor.json(), '"router-id": "10.113.63.240"')


if __name__ == '__main__':
    unittest.main()


# encoding: utf-8
"""
fragment.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.resource import BitResource


# =================================================================== Fragment

# Uses bitmask operand format defined above.
#   0   1   2   3   4   5   6   7
# +---+---+---+---+---+---+---+---+
# |   Reserved    |LF |FF |IsF|DF |
# +---+---+---+---+---+---+---+---+
#
# Bitmask values:
# +  Bit 7 - Don't fragment (DF)
# +  Bit 6 - Is a fragment (IsF)
# +  Bit 5 - First fragment (FF)
# +  Bit 4 - Last fragment (LF)

class Fragment (BitResource):
	NAME = 'fragment'

	NOT      = 0x00
	DONT     = 0x01
	IS       = 0x02
	FIRST    = 0x04
	LAST     = 0x08
	# reserved = 0xF0

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'NOT-A-FRAGMENT': NOT,
		'DONT-FRAGMENT':  DONT,
		'IS-FRAGMENT':    IS,
		'FIRST-FRAGMENT': FIRST,
		'LAST-FRAGMENT':  LAST,
	}.items())

	names = dict([(r,l) for (l,r) in codes.items()])



# encoding: utf-8
"""
mac_mobility.py

Created by Anton Aksola on 2018-11-03
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity

# ================================================================== MacMobility
# RFC 7432 Section 7.7.

@ExtendedCommunity.register
class MacMobility (ExtendedCommunity):
	COMMUNITY_TYPE = 0x06
	COMMUNITY_SUBTYPE = 0x00
	DESCRIPTION = 'mac-mobility'

	__slots__ = ['sequence','sticky']

	def __init__ (self, sequence, sticky=False, community=None):
		self.sequence = sequence
		self.sticky = sticky
		ExtendedCommunity.__init__(
			self,
			community if community else pack(
				'!2sBxI',
				self._subtype(transitive=True),
				1 if sticky else 0,
				sequence
			)
		)

	def __hash__ (self):
		return hash((self.sticky, self.sequence))

	def __repr__ (self):
		s = "%s:%d" % (self.DESCRIPTION, self.sequence)
		if self.sticky:
			s += ":sticky"
		return s

	@staticmethod
	def unpack (data):
		flags, seq = unpack('!BxI', data[2:8])
		return MacMobility(seq, True if flags == 1 else False)



# encoding: utf-8
"""
family.py

Created by Thomas Mangin on 2010-01-19.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.protocol.resource import Resource


# ======================================================================== AFI
# https://www.iana.org/assignments/address-family-numbers/

class _AFI (int):
	UNDEFINED = 0x00  # internal
	IPv4      = 0x01
	IPv6      = 0x02
	L2VPN     = 0x19
	BGPLS     = 0x4004

	_names = {
		UNDEFINED: 'undefined',
		IPv4:      'ipv4',
		IPv6:      'ipv6',
		L2VPN:     'l2vpn',
		BGPLS:     'bgp-ls',
	}

	_masks = {
		IPv4:  32,
		IPv6:  128,
	}

	def pack (self):
		return pack('!H',self)

	def name (self):
		return self._names.get(self,'unknown-afi-0x%s' % hex(self))

	def mask (self):
		return self._masks.get(self,'invalid request for this family')

	def __repr__ (self):
		return self.name()

	def __str__ (self):
		return self.name()


class AFI (Resource):
	undefined = _AFI(_AFI.UNDEFINED)
	ipv4      = _AFI(_AFI.IPv4)
	ipv6      = _AFI(_AFI.IPv6)
	l2vpn     = _AFI(_AFI.L2VPN)
	bgpls     = _AFI(_AFI.BGPLS)

	common = {
		undefined.pack(): undefined,
		ipv4.pack():      ipv4,
		ipv6.pack():      ipv6,
		l2vpn.pack():     l2vpn,
		bgpls.pack():     bgpls,
	}

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'ipv4':  ipv4,
		'ipv6':  ipv6,
		'l2vpn': l2vpn,
		'bgp-ls': bgpls,
	}.items())

	cache = dict([(r,r) for (l,r) in codes.items()])
	names = dict([(r,l) for (l,r) in codes.items()])

	inet_names = dict([(r,l.replace('ipv','inet')) for (l,r) in codes.items()])

	def name (self):
		return self.inet_names.get(self,"unknown afi")

	@staticmethod
	def unpack (data):
		return AFI.common.get(data,_AFI(unpack('!H',data)[0]))

	@classmethod
	def value (cls,name):
		return cls.codes.get(name,None)

	@staticmethod
	def implemented_safi (afi):
		if afi == 'ipv4':
			return ['unicast','multicast','nlri-mpls','mpls-vpn','flow','flow-vpn']
		if afi == 'ipv6':
			return ['unicast','mpls-vpn','flow','flow-vpn']
		if afi == 'l2vpn':
			return ['vpls','evpn']
		if afi == 'bgp-ls':
			return ['bgp-ls','bgp-ls-vpn']
		return []

	@classmethod
	def fromString (cls, string):
		return cls.codes.get(string,cls.undefined)

	@classmethod
	def create (cls, value):
		return cls.cache.get(value,_AFI(value))


# ======================================================================= SAFI
# https://www.iana.org/assignments/safi-namespace

class _SAFI (int):
	UNDEFINED  = 0              # internal
	UNICAST    = 1              # [RFC4760]
	MULTICAST  = 2              # [RFC4760]
	NLRI_MPLS  = 4              # [RFC3107]
	VPLS       = 65             # [RFC4761]
	EVPN       = 70             # [draft-ietf-l2vpn-evpn]
	BGPLS      = 71             # [RFC7752]
	BGPLS_VPN  = 72             # [RFC7752]
	MPLS_VPN   = 128            # [RFC4364]
	RTC        = 132            # [RFC4684]
	FLOW_IP    = 133            # [RFC5575]
	FLOW_VPN   = 134            # [RFC5575]
	# deprecated = 3            # [RFC4760]
	# mcast_vpn = 5             # [draft-ietf-l3vpn-2547bis-mcast-bgp] (TEMPORARY - Expires 2008-06-19)
	# pseudowire = 6            # [draft-ietf-pwe3-dynamic-ms-pw] (TEMPORARY - Expires 2008-08-23) Dynamic Placement of Multi-Segment Pseudowires
	# encapsulation = 7         # [RFC5512]
	# tunel = 64                # [Nalawade]
	# bgp_mdt = 66              # [Nalawade]
	# bgp_4over6 = 67           # [Cui]
	# bgp_6over4 = 67           # [Cui]
	# vpn_adi = 69              # [RFC-ietf-l1vpn-bgp-auto-discovery-05.txt]
	# mcast_bgp_mpls_vpn = 129  # [RFC2547]
	# rt = 132                  # [RFC4684]
	# vpn_ad = 140              # [draft-ietf-l3vpn-bgpvpn-auto]
	# private = [_ for _ in range(241,254)]   # [RFC4760]
	# unassigned = [_ for _ in range(8,64)] + [_ for _ in range(70,128)]
	# reverved = [0,3] + [130,131] + [_ for _ in range(135,140)] + [_ for _ in range(141,241)] + [255,]    # [RFC4760]

	_names = {
		UNICAST:   'unicast',
		MULTICAST: 'multicast',
		NLRI_MPLS: 'nlri-mpls',
		VPLS:      'vpls',
		EVPN:      'evpn',
		BGPLS:     'bgp-ls',
		BGPLS_VPN: 'bgp-ls-vpn',
		MPLS_VPN:  'mpls-vpn',
		RTC:       'rtc',
		FLOW_IP:   'flow',
		FLOW_VPN:  'flow-vpn',
	}

	def pack (self):
		return character(self)

	def name (self):
		return self._names.get(self,'unknown safi %d' % int(self))

	def has_label (self):
		return self in (SAFI.nlri_mpls,SAFI.mpls_vpn)

	def has_rd (self):
		return self in (SAFI.nlri_mpls,SAFI.mpls_vpn,SAFI.flow_vpn)
		# technically self.flow_vpn and self.vpls has an RD but it is not an NLRI

	def has_path (self):
		return self in (SAFI.unicast,SAFI.nlri_mpls)
		# technically self.flow_vpn and self.vpls has an RD but it is not an NLRI

	def __str__ (self):
		return self.name()

	def __repr__ (self):
		return str(self)


class SAFI (Resource):
	undefined  = _SAFI(_SAFI.UNDEFINED)
	unicast    = _SAFI(_SAFI.UNICAST)
	multicast  = _SAFI(_SAFI.MULTICAST)
	nlri_mpls  = _SAFI(_SAFI.NLRI_MPLS)
	vpls       = _SAFI(_SAFI.VPLS)
	evpn       = _SAFI(_SAFI.EVPN)
	bgp_ls     = _SAFI(_SAFI.BGPLS)
	bgp_ls_vpn = _SAFI(_SAFI.BGPLS_VPN)
	mpls_vpn   = _SAFI(_SAFI.MPLS_VPN)
	rtc        = _SAFI(_SAFI.RTC)
	flow_ip    = _SAFI(_SAFI.FLOW_IP)
	flow_vpn   = _SAFI(_SAFI.FLOW_VPN)

	common = {
		undefined.pack():  undefined,
		unicast.pack():    unicast,
		multicast.pack():  multicast,
		nlri_mpls.pack():  nlri_mpls,
		vpls.pack():       vpls,
		evpn.pack():       evpn,
		bgp_ls.pack():     bgp_ls,
		bgp_ls_vpn.pack(): bgp_ls_vpn,
		mpls_vpn.pack():   mpls_vpn,
		rtc.pack():        rtc,
		flow_ip.pack():    flow_ip,
		flow_vpn.pack():   flow_vpn,
	}

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'unicast':   unicast,
		'multicast': multicast,
		'nlri-mpls': nlri_mpls,
		'vpls':      vpls,
		'evpn':      evpn,
		'bgp-ls':    bgp_ls,
		'bgp-ls-vpn':bgp_ls_vpn,
		'mpls-vpn':  mpls_vpn,
		'rtc':       rtc,
		'flow':      flow_ip,
		'flow-vpn':  flow_vpn,
	}.items())

	names = _SAFI._names

	cache = dict([(r,r) for (l,r) in codes.items()])

	@staticmethod
	def unpack (data):
		return SAFI.common.get(data,_SAFI(ordinal(data)))

	@classmethod
	def value (cls,name):
		return cls.codes.get(name,None)

	@classmethod
	def fromString (cls, string):
		return cls.codes.get(string,cls.undefined)

	@classmethod
	def create (cls, value):
		return cls.cache.get(value,_SAFI(value))


# ===================================================================== FAMILY


class Family (object):
	size = {
		# family                   next-hop   RD
		(AFI.ipv4,SAFI.unicast):   ((4,),      0),
		(AFI.ipv4,SAFI.multicast): ((4,),      0),
		(AFI.ipv4,SAFI.nlri_mpls): ((4,),      0),
		(AFI.ipv4,SAFI.mpls_vpn):  ((12,),     8),
		(AFI.ipv4,SAFI.flow_ip):   ((0,4),     0),
		(AFI.ipv4,SAFI.flow_vpn):  ((0,4),     0),
		(AFI.ipv4,SAFI.rtc):       ((4,16),    0),
		(AFI.ipv6,SAFI.unicast):   ((16,32),   0),
		(AFI.ipv6,SAFI.nlri_mpls): ((16,32),   0),
		(AFI.ipv6,SAFI.mpls_vpn):  ((24,40),   8),
		(AFI.ipv6,SAFI.flow_ip):   ((0,16,32), 0),
		(AFI.ipv6,SAFI.flow_vpn):  ((0,16,32), 0),
		(AFI.l2vpn,SAFI.vpls):     ((4,),      0),
		(AFI.l2vpn,SAFI.evpn):     ((4,),      0),
		(AFI.bgpls,SAFI.bgp_ls):   ((4,),      0),
	}

	__slots__ = ['afi','safi']

	def __init__ (self, afi, safi):
		self.afi = AFI.create(afi)
		self.safi = SAFI.create(safi)

	def has_label (self):
		return self.safi.has_label()

	def has_rd (self):
		return self.safi.has_rd()

	def has_path (self):
		return self.safi.has_path()

	def __eq__ (self, other):
		return \
			self.afi == other.afi and \
			self.safi == other.safi

	def __neq__ (self, other):
		return \
			self.afi != other.afi or \
			self.safi != other.safi

	def __lt__ (self, other):
		raise RuntimeError('comparing Family for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing Family for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing Family for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing Family for ordering does not make sense')

	def family (self):
		return (self.afi,self.safi)

	def extensive (self):
		return 'afi %s safi %s' % (self.afi,self.safi)

	def __repr__ (self):
		return "%s %s" % (str(self.afi),str(self.safi))



# encoding: utf-8
"""
timer.py

Created by Thomas Mangin on 2012-07-21.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time

from exabgp.logger import Logger
from exabgp.bgp.message import _NOP
from exabgp.bgp.message import KeepAlive
from exabgp.bgp.message import Notify

# ================================================================ ReceiveTimer
# Track the time for keepalive updates


class ReceiveTimer (object):
	def __init__ (self, session, holdtime, code, subcode, message=''):
		self.logger = Logger()
		self.session = session

		self.holdtime = holdtime
		self.last_print = 0
		self.last_read = int(time.time())

		self.code = code
		self.subcode = subcode
		self.message = message
		self.single = False

	def check_ka_timer (self, message=_NOP,ignore=_NOP.TYPE):
		if self.holdtime == 0:
			return message.TYPE != KeepAlive.TYPE
		now = int(time.time())
		if message.TYPE != ignore:
			self.last_read = now
		elapsed = now - self.last_read
		if elapsed > self.holdtime:
			raise Notify(self.code,self.subcode,self.message)
		if self.last_print != now:
			left = self.holdtime - elapsed
			self.logger.debug('receive-timer %d second(s) left' % left,source='ka-'+self.session())
			self.last_print = now
		return True

	def check_ka (self, message=_NOP,ignore=_NOP.TYPE):
		if self.check_ka_timer(message,ignore):
			return
		if self.single:
			raise Notify(2,6,'Negotiated holdtime was zero, it was invalid to send us a keepalive messages')
		self.single = True


class SendTimer (object):
	def __init__ (self, session, holdtime):
		self.logger = Logger()
		self.session = session

		self.keepalive = holdtime.keepalive()
		self.last_print = int(time.time())
		self.last_sent = int(time.time())

	def need_ka (self):
		if not self.keepalive:
			return False

		now  = int(time.time())
		left = self.last_sent + self.keepalive - now

		if now != self.last_print:
			self.logger.debug('send-timer %d second(s) left' % left,source='ka-'+self.session())
			self.last_print = now

		if left <= 0:
			self.last_sent = now
			return True
		return False



# encoding: utf-8
"""
admingroup.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE


@LINKSTATE.register()
class AdminGroup(object):
	TLV = 1088

	def __init__ (self, colormask):
		self.colormask = colormask

	def __repr__ (self):
		return "Admin Group mask: %s" % (self.colormask)

	@classmethod
	def unpack (cls,data,length):
		if length != 4:
			raise Notify(3,5, "Unable to decode attribute. Incorrect Size")
		else:
			b = BitArray(bytes=data)
			colormask = b.unpack('uintbe:32')
			return cls(colormask=colormask)

	def json (self,compact=None):
		return '"admin-group-mask": %s' % self.colormask



# encoding: utf-8
"""
nop.py

Created by Thomas Mangin on 2009-11-06.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character

from exabgp.bgp.message.message import Message

# ========================================================================= NOP
#


class NOP (Message):
	ID = Message.CODE.NOP
	TYPE = character(Message.CODE.NOP)

	def message (self,negotiated=None):
		raise RuntimeError('NOP messages can not be sent on the wire')

	def __str__ (self):
		return "NOP"

	@classmethod
	def unpack_message (cls, data, negotiated):  # pylint: disable=W0613
		return NOP()

_NOP = NOP()



# encoding: utf-8
"""
capability.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys

# Do not create a dependency loop by using exabgp.bgp.message as import
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i

from exabgp.bgp.message.notification import Notify


class _CapabilityCode (int):
	_cache = dict()

	if sys.version_info[0] < 3:
		__slots__ = ['NAME','_cache']

	RESERVED                 = 0x00  # [RFC5492]
	MULTIPROTOCOL            = 0x01  # [RFC2858]
	ROUTE_REFRESH            = 0x02  # [RFC2918]
	OUTBOUND_ROUTE_FILTERING = 0x03  # [RFC5291]
	MULTIPLE_ROUTES          = 0x04  # [RFC3107]
	EXTENDED_NEXT_HOP        = 0x05  # [RFC5549]
	EXTENDED_MESSAGE         = 0x06  # https://tools.ietf.org/html/draft-ietf-idr-bgp-extended-messages-24

	# 6-63      Unassigned
	GRACEFUL_RESTART         = 0x40  # [RFC4724]
	FOUR_BYTES_ASN           = 0x41  # [RFC4893]
	# 66        Deprecated
	DYNAMIC_CAPABILITY       = 0x43  # [Chen]
	MULTISESSION             = 0x44  # [draft-ietf-idr-bgp-multisession]
	ADD_PATH                 = 0x45  # [draft-ietf-idr-add-paths]
	ENHANCED_ROUTE_REFRESH   = 0x46  # [draft-ietf-idr-bgp-enhanced-route-refresh]
	# 70-127    Unassigned
	ROUTE_REFRESH_CISCO      = 0x80  # I Can only find reference to this in the router logs
	# 128-255   Reserved for Private Use [RFC5492]
	MULTISESSION_CISCO       = 0x83  # What Cisco really use for Multisession (yes this is a reserved range in prod !)

	HOSTNAME                 = 0xB8  # ExaBGP only ...
	OPERATIONAL              = 0xB9  # ExaBGP only ...

	# Internal
	AIGP = 0xFF00

	names = {
		RESERVED:                  'reserved',
		MULTIPROTOCOL:             'multiprotocol',
		ROUTE_REFRESH:             'route-refresh',
		OUTBOUND_ROUTE_FILTERING:  'outbound-route-filtering',
		MULTIPLE_ROUTES:           'multiple-routes',
		EXTENDED_NEXT_HOP:         'extended-next-hop',
		EXTENDED_MESSAGE:          'extended-message',

		GRACEFUL_RESTART:          'graceful-restart',
		FOUR_BYTES_ASN:            'asn4',

		DYNAMIC_CAPABILITY:        'dynamic-capability',
		MULTISESSION:              'multi-session',
		ADD_PATH:                  'add-path',
		ENHANCED_ROUTE_REFRESH:    'enhanced-route-refresh',
		OPERATIONAL:               'operational',

		ROUTE_REFRESH_CISCO:       'cisco-route-refresh',
		MULTISESSION_CISCO:        'cisco-multi-sesion',

		AIGP:                      'aigp',

		HOSTNAME:                  'exabgp-experimental-hostname',
		OPERATIONAL:               'exabgp-experimental-operational',
	}

	def __new__ (cls, value):
		if value in cls._cache:
			return cls._cache[value]
		obj = super(_CapabilityCode, cls).__new__(cls,value)
		obj.NAME = cls.names.get(value,'unknown capability %s' % hex(value))
		cls._cache[value] = obj
		return obj

	def __str__ (self):
		return self.names.get(self,'unknown capability %s' % hex(self))

	def __repr__ (self):
		return str(self)

	def name (self):
		return self.names.get(self,'unknown capability %s' % hex(self))


# =================================================================== Capability
#

class Capability (object):

	class CODE (int):
		if sys.version_info[0] < 3:
			__slots__ = []

		RESERVED                 = _CapabilityCode(_CapabilityCode.RESERVED)
		MULTIPROTOCOL            = _CapabilityCode(_CapabilityCode.MULTIPROTOCOL)
		ROUTE_REFRESH            = _CapabilityCode(_CapabilityCode.ROUTE_REFRESH)
		OUTBOUND_ROUTE_FILTERING = _CapabilityCode(_CapabilityCode.OUTBOUND_ROUTE_FILTERING)
		MULTIPLE_ROUTES          = _CapabilityCode(_CapabilityCode.MULTIPLE_ROUTES)
		EXTENDED_NEXT_HOP        = _CapabilityCode(_CapabilityCode.EXTENDED_NEXT_HOP)
		EXTENDED_MESSAGE         = _CapabilityCode(_CapabilityCode.EXTENDED_MESSAGE)
		GRACEFUL_RESTART         = _CapabilityCode(_CapabilityCode.GRACEFUL_RESTART)
		FOUR_BYTES_ASN           = _CapabilityCode(_CapabilityCode.FOUR_BYTES_ASN)
		DYNAMIC_CAPABILITY       = _CapabilityCode(_CapabilityCode.DYNAMIC_CAPABILITY)
		MULTISESSION             = _CapabilityCode(_CapabilityCode.MULTISESSION)
		ADD_PATH                 = _CapabilityCode(_CapabilityCode.ADD_PATH)
		ENHANCED_ROUTE_REFRESH   = _CapabilityCode(_CapabilityCode.ENHANCED_ROUTE_REFRESH)
		ROUTE_REFRESH_CISCO      = _CapabilityCode(_CapabilityCode.ROUTE_REFRESH_CISCO)
		MULTISESSION_CISCO       = _CapabilityCode(_CapabilityCode.MULTISESSION_CISCO)
		HOSTNAME                 = _CapabilityCode(_CapabilityCode.HOSTNAME)
		OPERATIONAL              = _CapabilityCode(_CapabilityCode.OPERATIONAL)
		AIGP                     = _CapabilityCode(_CapabilityCode.AIGP)

		unassigned = range(70,128)
		reserved = range(128,256)

		def __str__ (self):
			name = _CapabilityCode.names.get(self,None)
			if name is None:
				if self in Capability.CODE.unassigned:
					return 'unassigned-%s' % hex(self)
				if self in Capability.CODE.reserved:
					return 'reserved-%s' % hex(self)
				return 'capability-%s' % hex(self)
			return name

		def __repr__ (self):
			return str(self)

		# XXX: Could use cls instead of _CapabilityCode and other tidy up
		@classmethod
		def name (cls, self):
			name = _CapabilityCode.names.get(self,None)
			if name is None:
				if self in Capability.CODE.unassigned:
					return 'unassigned-%s' % hex(self)
				if self in Capability.CODE.reserved:
					return 'reserved-%s' % hex(self)
			return name

	registered_capability = dict()
	unknown_capability = None

	@staticmethod
	def hex (data):
		return '0x' + concat_bytes_i('%02x' % ordinal(_) for _ in data)

	@classmethod
	def unknown (cls, klass):
		if cls.unknown_capability is not None:
			raise RuntimeError('only one fallback function can be registered')
		cls.unknown_capability = klass

	@classmethod
	def register (cls, capability=None):
		def register_capability (klass):
			# ID is defined by all the subclasses - otherwise they do not work :)
			what = klass.ID if capability is None else capability  # pylint: disable=E1101
			if what in cls.registered_capability:
				raise RuntimeError('only one class can be registered per capability')
			cls.registered_capability[what] = klass
			return klass
		return register_capability

	@classmethod
	def klass (cls, what):
		if what in cls.registered_capability:
			kls = cls.registered_capability[what]
			kls.ID = what
			return kls
		if cls.unknown_capability:
			return cls.unknown_capability
		raise Notify (2,4,'can not handle capability %s' % what)

	@classmethod
	def unpack (cls, capability, capabilities, data):
		instance = capabilities.get(capability,Capability.klass(capability)())
		return cls.klass(capability).unpack_capability(instance,data,capability)



# encoding: utf-8
"""
reactor/loop.py

Created by Thomas Mangin on 2012-06-10.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time
import uuid
import select
import socket

from exabgp.util import character
from exabgp.util import concat_bytes_i

from exabgp.reactor.daemon import Daemon
from exabgp.reactor.listener import Listener
from exabgp.reactor.api.processes import Processes
from exabgp.reactor.api.processes import ProcessError
from exabgp.reactor.peer import Peer
from exabgp.reactor.peer import ACTION
from exabgp.reactor.asynchronous import ASYNC
from exabgp.reactor.interrupt import Signal
from exabgp.reactor.network.error import error

from exabgp.reactor.api import API
from exabgp.configuration.configuration import Configuration
from exabgp.configuration.environment import environment

from exabgp.version import version
from exabgp.logger import Logger


class Reactor (object):
	class Exit (object):
		normal = 0
		validate = 0
		listening = 1
		configuration = 1
		privileges = 1
		log = 1
		pid = 1
		socket = 1
		io_error = 1
		process = 1
		select = 1
		unknown = 1

	# [hex(ord(c)) for c in os.popen('clear').read()]
	clear = concat_bytes_i(character(int(c,16)) for c in ['0x1b', '0x5b', '0x48', '0x1b', '0x5b', '0x32', '0x4a'])

	def __init__ (self, configurations):
		self._ips = environment.settings().tcp.bind
		self._port = environment.settings().tcp.port
		self._stopping = environment.settings().tcp.once
		self.exit_code = self.Exit.unknown

		self.max_loop_time = environment.settings().reactor.speed
		self._sleep_time = self.max_loop_time / 100
		self._busyspin = {}
		self.early_drop = environment.settings().daemon.drop

		self.processes = None

		self.configuration = Configuration(configurations)
		self.logger = Logger()
		self.asynchronous = ASYNC()
		self.signal = Signal()
		self.daemon = Daemon(self)
		self.listener = Listener(self)
		self.api = API(self)

		self.peers = {}

		self._reload_processes = False
		self._saved_pid = False

	def _termination (self,reason, exit_code):
		self.exit_code = exit_code
		self.signal.received = Signal.SHUTDOWN
		self.logger.critical(reason,'reactor')

	def _prevent_spin(self):
		second = int(time.time())
		if not second in self._busyspin:
			self._busyspin = {second: 0}
		self._busyspin[second] += 1
		if self._busyspin[second] > self.max_loop_time:
			time.sleep(self._sleep_time)
			return True
		return False

	def _api_ready (self,sockets,sleeptime):
		fds = self.processes.fds()
		ios = fds + sockets

		try:
			read,_,_ = select.select(ios,[],[],sleeptime)
			for fd in fds:
				if fd in read:
					read.remove(fd)
			return read
		except select.error as exc:
			err_no,message = exc.args  # pylint: disable=W0633
			if err_no not in error.block:
				raise exc
			self._prevent_spin()
			return []
		except socket.error as exc:
			# python 3 does not raise on closed FD, but python2 does
			# we have lost a peer and it is causing the select
			# to complain, the code will self-heal, ignore the issue
			# (EBADF from python2 must be ignored if when checkign error.fatal)
			# otherwise sending  notification causes TCP to drop and cause
			# this code to kill ExaBGP
			self._prevent_spin()
			return []
		except ValueError as exc:
			# The peer closing the TCP connection lead to a negative file descritor
			self._prevent_spin()
			return []
		except KeyboardInterrupt:
			self._termination('^C received',self.Exit.normal)
			return []

	def _active_peers (self):
		peers = set()
		for key,peer in self.peers.items():
			if not peer.neighbor.passive or peer.proto:
				peers.add(key)
		return peers

	def _completed (self,peers):
		for peer in peers:
			if self.peers[peer].neighbor.rib.outgoing.pending():
				return False
		return True

	def run (self, validate, root):
		self.daemon.daemonise()

		# Make sure we create processes once we have closed file descriptor
		# unfortunately, this must be done before reading the configuration file
		# so we can not do it with dropped privileges
		self.processes = Processes()

		# we have to read the configuration possibly with root privileges
		# as we need the MD5 information when we bind, and root is needed
		# to bind to a port < 1024

		# this is undesirable as :
		# - handling user generated data as root should be avoided
		# - we may not be able to reload the configuration once the privileges are dropped

		# but I can not see any way to avoid it
		for ip in self._ips:
			if not self.listener.listen_on(ip, None, self._port, None, False, None):
				return self.Exit.listening

		if not self.load():
			return self.Exit.configuration

		if validate:  # only validate configuration
			self.logger.warning('','configuration')
			self.logger.warning('parsed Neighbors, un-templated','configuration')
			self.logger.warning('------------------------------','configuration')
			self.logger.warning('','configuration')
			for key in self.peers:
				self.logger.warning(str(self.peers[key].neighbor),'configuration')
				self.logger.warning('','configuration')
			return self.Exit.validate

		for neighbor in self.configuration.neighbors.values():
			if neighbor.listen:
				if not self.listener.listen_on(neighbor.md5_ip, neighbor.peer_address, neighbor.listen, neighbor.md5_password, neighbor.md5_base64, neighbor.ttl_in):
					return self.Exit.listening

		if not self.early_drop:
			self.processes.start(self.configuration.processes)

		if not self.daemon.drop_privileges():
			self.logger.critical('could not drop privileges to \'%s\' refusing to run as root' % self.daemon.user,'reactor')
			self.logger.critical('set the environmemnt value exabgp.daemon.user to change the unprivileged user','reactor')
			return self.Exit.privileges

		if self.early_drop:
			self.processes.start(self.configuration.processes)

		# This is required to make sure we can write in the log location as we now have dropped root privileges
		if not self.logger.restart():
			self.logger.critical('could not setup the logger, aborting','reactor')
			return self.Exit.log

		if not self.daemon.savepid():
			return self.Exit.pid

		# did we complete the run of updates caused by the last SIGUSR1/SIGUSR2 ?
		reload_completed = False

		wait = environment.settings().tcp.delay
		if wait:
			sleeptime = (wait * 60) - int(time.time()) % (wait * 60)
			self.logger.debug('waiting for %d seconds before connecting' % sleeptime,'reactor')
			time.sleep(float(sleeptime))

		workers = {}
		peers = set()

		while True:
			try:
				if self.signal.received:
					for key in self.peers:
						if self.peers[key].neighbor.api['signal']:
							self.peers[key].reactor.processes.signal(self.peers[key].neighbor,self.signal.number)

					signaled = self.signal.received
					self.signal.rearm()

					if signaled == Signal.SHUTDOWN:
						self.shutdown()
						break

					if signaled == Signal.RESTART:
						self.restart()
						continue

					if not reload_completed:
						continue

					if signaled == Signal.FULL_RELOAD:
						self._reload_processes = True

					if signaled in (Signal.RELOAD, Signal.FULL_RELOAD):
						self.load()
						self.processes.start(self.configuration.processes,self._reload_processes)
						self._reload_processes = False
						continue

				if self.listener.incoming():
					# check all incoming connection
					self.asynchronous.schedule(str(uuid.uuid1()),'checking for new connection(s)',self.listener.new_connections())

				peers = self._active_peers()
				if self._completed(peers):
					reload_completed = True

				sleep = self._sleep_time

				# do not attempt to listen on closed sockets even if the peer is still here
				for io in list(workers.keys()):
					try:
						if io.fileno() == -1:
							del workers[io]
					except socket.error:
						del workers[io]

				# give a turn to all the peers
				for key in list(peers):
					peer = self.peers[key]
					action = peer.run()

					# .run() returns an ACTION enum:
					# * immediate if it wants to be called again
					# * later if it should be called again but has no work atm
					# * close if it is finished and is closing down, or restarting
					if action == ACTION.CLOSE:
						if key in self.peers:
							del self.peers[key]
						peers.discard(key)
					# we are loosing this peer, not point to schedule more process work
					elif action == ACTION.LATER:
						for io in peer.sockets():
							workers[io] = key
						# no need to come back to it before a a full cycle
						peers.discard(key)
					elif action == ACTION.NOW:
						sleep = 0

					if not peers:
						break

				# read at least on message per process if there is some and parse it
				for service,command in self.processes.received():
					self.api.text(self,service,command)
					sleep = 0

				self.asynchronous.run()

				for io in self._api_ready(list(workers),sleep):
					peers.add(workers[io])
					del workers[io]

				if self._stopping and not self.peers.keys():
					self._termination('exiting on peer termination',self.Exit.normal)

			except KeyboardInterrupt:
				self._termination('^C received',self.Exit.normal)
			except SystemExit:
				self._termination('exiting', self.Exit.normal)
			# socket.error is a subclass of IOError (so catch it first)
			except socket.error:
				self._termination('socket error received',self.Exit.socket)
			except IOError:
				self._termination('I/O Error received, most likely ^C during IO',self.Exit.io_error)
			except ProcessError:
				self._termination('Problem when sending message(s) to helper program, stopping',self.Exit.process)
			except select.error:
				self._termination('problem using select, stopping',self.Exit.select)

		return self.exit_code

	def shutdown (self):
		"""Terminate all the current BGP connections"""
		self.logger.critical('performing shutdown','reactor')
		if self.listener:
			self.listener.stop()
			self.listener = None
		for key in self.peers.keys():
			self.peers[key].shutdown()
		self.asynchronous.clear()
		self.processes.terminate()
		self.daemon.removepid()
		self._stopping = True

	def load (self):
		"""Reload the configuration and send to the peer the route which changed"""
		self.logger.notice('performing reload of exabgp %s' % version,'configuration')

		reloaded = self.configuration.reload()

		if not reloaded:
			#
			# Careful the string below is used but the QA code to check for sucess of failure
			self.logger.error('problem with the configuration file, no change done','configuration')
			# Careful the string above is used but the QA code to check for sucess of failure
			#
			self.logger.error(str(self.configuration.error),'configuration')
			return False

		for key, peer in self.peers.items():
			if key not in self.configuration.neighbors:
				self.logger.debug('removing peer: %s' % peer.neighbor.name(),'reactor')
				peer.remove()

		for key, neighbor in self.configuration.neighbors.items():
			# new peer
			if key not in self.peers:
				self.logger.debug('new peer: %s' % neighbor.name(),'reactor')
				peer = Peer(neighbor,self)
				self.peers[key] = peer
			# modified peer
			elif self.peers[key].neighbor != neighbor:
				self.logger.debug('peer definition change, establishing a new connection for %s' % str(key),'reactor')
				self.peers[key].reestablish(neighbor)
			# same peer but perhaps not the routes
			else:
				# finding what route changed and sending the delta is not obvious
				self.logger.debug('peer definition identical, updating peer routes if required for %s' % str(key),'reactor')
				self.peers[key].reconfigure(neighbor)
			for ip in self._ips:
				if ip.afi == neighbor.peer_address.afi:
					self.listener.listen_on(ip, neighbor.peer_address, self._port, neighbor.md5_password, neighbor.md5_base64, None)
		self.logger.notice('loaded new configuration successfully','reactor')

		return True

	def restart (self):
		"""Kill the BGP session and restart it"""
		self.logger.notice('performing restart of exabgp %s' % version,'reactor')
		self.configuration.reload()

		for key in self.peers.keys():
			if key not in self.configuration.neighbors.keys():
				neighbor = self.configuration.neighbors[key]
				self.logger.debug('removing Peer %s' % neighbor.name(),'reactor')
				self.peers[key].remove()
			else:
				self.peers[key].reestablish()
		self.processes.start(self.configuration.processes,True)

	# def nexthops (self, peers):
	# 	return dict((peer,self.peers[peer].neighbor.local_address) for peer in peers)



# encoding: utf-8
"""
family.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.update.nlri.flow import NLRI

from exabgp.configuration.core import Section


class ParseFamily (Section):
	syntax = \
		'family {\n' \
		'   all;      # default if not family block is present, announce all we know\n' \
		'   \n' \
		'   ipv4 unicast;\n' \
		'   ipv4 multicast;\n' \
		'   ipv4 nlri-mpls;\n' \
		'   ipv4 mpls-vpn;\n' \
		'   ipv4 flow;\n' \
		'   ipv4 flow-vpn;\n' \
		'   ipv6 unicast;\n' \
		'   ipv6 flow;\n' \
		'   ipv6 flow-vpn;\n' \
		'   l2vpn vpls;\n' \
		'   l2vpn evpn;\n' \
		'}'

	convert = {
		'ipv4': {
			'unicast':   	(AFI.ipv4,SAFI.unicast),
			'multicast': 	(AFI.ipv4,SAFI.multicast),
			'nlri-mpls': 	(AFI.ipv4,SAFI.nlri_mpls),
			'mpls-vpn':  	(AFI.ipv4,SAFI.mpls_vpn),
			'flow':      	(AFI.ipv4,SAFI.flow_ip),
			'flow-vpn':  	(AFI.ipv4,SAFI.flow_vpn),
		},
		'ipv6': {
			'unicast':   	(AFI.ipv6,SAFI.unicast),
			'nlri-mpls': 	(AFI.ipv6,SAFI.nlri_mpls),
			'mpls-vpn':  	(AFI.ipv6,SAFI.mpls_vpn),
			'flow':      	(AFI.ipv6,SAFI.flow_ip),
			'flow-vpn':  	(AFI.ipv6,SAFI.flow_vpn),
		},
		'l2vpn': {
			'vpls':      	(AFI.l2vpn,SAFI.vpls),
			'evpn':      	(AFI.l2vpn,SAFI.evpn),
		},
		'bgp-ls': {
			'bgp-ls':		(AFI.bgpls,SAFI.bgp_ls),
			'bgp-ls-vpn':	(AFI.bgpls,SAFI.bgp_ls_vpn),
		},
	}

	action = {
		'ipv4':  'append-command',
		'ipv6':  'append-command',
		'l2vpn': 'append-command',
		'bgp-ls': 'append-command',
	}

	name = 'family'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)
		self.known = {
			'ipv4':  self.ipv4,
			'ipv6':  self.ipv6,
			'l2vpn': self.l2vpn,
			'bgp-ls': self.bgpls,
		}
		self._all = ''
		self._seen = []

	def clear (self):
		self._all = False
		self._seen = []

	def pre (self):
		self.clear()
		return True

	def post (self):
		return True

	def _family (self, tokeniser, afi):
		if self._all:
			raise ValueError('can not add any family once family all is set')

		safi = tokeniser().lower()

		pair = self.convert[afi].get(safi,None)
		if not pair:
			raise ValueError('invalid afi/safi pair %s/%s' % (afi,safi))
		if pair in self._seen:
			raise ValueError('duplicate afi/safi pair %s/%s' % (afi,safi))
		self._seen.append(pair)
		return pair

	def ipv4 (self, tokeniser):
		return self._family(tokeniser, 'ipv4')

	def ipv6 (self, tokeniser):
		return self._family(tokeniser, 'ipv6')

	def l2vpn (self, tokeniser):
		return self._family(tokeniser, 'l2vpn')

	def bgpls (self, tokeniser):
		return self._family(tokeniser, 'bgp-ls')

	def minimal (self, tokeniser):
		raise ValueError('family minimal is deprecated')

	def all (self, tokeniser):
		if self._all or self._seen:
			return self.error.set('all can not be used with any other options')
		self._all = True
		for pair in NLRI.known_families():
			self._seen.append(pair)


class ParseAddPath (ParseFamily):
	name = 'add-path'



# encoding: utf-8
"""
origin.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute.attribute import Attribute


# =================================================================== Origin (1)

@Attribute.register()
class Origin (Attribute):
	ID = Attribute.CODE.ORIGIN
	FLAG = Attribute.Flag.TRANSITIVE
	CACHING = True

	IGP        = 0x00
	EGP        = 0x01
	INCOMPLETE = 0x02

	__slots__ = ['origin','_packed']

	def __init__ (self, origin, packed=None):
		self.origin = origin
		self._packed = self._attribute(packed if packed else character(origin))

	def __eq__ (self, other):
		return \
			self.ID == other.ID and \
			self.FLAG == other.FLAG and \
			self.origin == other.origin

	def __ne__ (self, other):
		return not self.__eq__(other)

	def pack (self, negotiated=None):
		return self._packed

	def __len__ (self):
		return len(self._packed)

	def __repr__ (self):
		if self.origin == 0x00:
			return 'igp'
		if self.origin == 0x01:
			return 'egp'
		if self.origin == 0x02:
			return 'incomplete'
		return 'invalid'

	@classmethod
	def unpack (cls, data, negotiated):
		return cls(ordinal(data),data)

	@classmethod
	def setCache (cls):
		# there can only be three, build them now
		IGP = Origin(Origin.IGP)
		EGP = Origin(Origin.EGP)
		INC = Origin(Origin.INCOMPLETE)

		cls.cache[Attribute.CODE.ORIGIN][IGP.pack()] = IGP
		cls.cache[Attribute.CODE.ORIGIN][EGP.pack()] = EGP
		cls.cache[Attribute.CODE.ORIGIN][INC.pack()] = INC


Origin.setCache()



# encoding: utf-8
"""
sr/labelindex.py

Created by Evelio Vila 2017-02-16
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""
from struct import pack, unpack

from exabgp.util import concat_bytes
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.sr.prefixsid import PrefixSid

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |       Type    |             Length            |   RESERVED    |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |            Flags              |       Label Index             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |          Label Index          |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 3.1.  Label-Index TLV


@PrefixSid.register()
class SrLabelIndex(object):
	TLV = 1
	LENGTH = 7

	def __init__ (self, labelindex, packed=None):
		self.labelindex = labelindex
		self.packed = self.pack()

	def __repr__ (self):
		return "%s" % (self.labelindex)

	def pack (self):
		return concat_bytes(
			pack('!B', self.TLV),
			pack('!H', self.LENGTH),
			pack('!B',0),  # reserved
			pack('!H',0),  # flags
			pack('!I',self.labelindex)
		)

	@classmethod
	def unpack (cls,data,length):
		labelindex = -1
		if cls.LENGTH != length:
			raise Notify(3,5, "Invalid TLV size. Should be 7 but {0} received".format(length))
		# Shift reserved bits
		data = data[1:]
		# Shift Flags
		# Flags: 16 bits of flags.  None is defined by this document.  The
		# flag field MUST be clear on transmission and MUST be ignored at
		# reception.
		data = data[2:6]
		labelindex = unpack('!I',data)[0]
		return cls(labelindex=labelindex,packed=data)

	def json (self,compact=None):
		return '"sr-label-index": %d' % (self.labelindex)



# encoding: utf-8
"""
configuration.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys

from exabgp.vendoring import six

from exabgp.logger import Logger

from exabgp.configuration.core import Error
from exabgp.configuration.core import Scope
from exabgp.configuration.core import Tokeniser
from exabgp.configuration.core import Section

from exabgp.configuration.process import ParseProcess
from exabgp.configuration.template import ParseTemplate
from exabgp.configuration.template.neighbor import ParseTemplateNeighbor
from exabgp.configuration.neighbor import ParseNeighbor
from exabgp.configuration.neighbor.api import ParseAPI
from exabgp.configuration.neighbor.api import ParseSend
from exabgp.configuration.neighbor.api import ParseReceive
from exabgp.configuration.neighbor.family import ParseFamily
from exabgp.configuration.neighbor.family import ParseAddPath
from exabgp.configuration.capability import ParseCapability
from exabgp.configuration.announce import SectionAnnounce
from exabgp.configuration.announce import AnnounceIPv4
from exabgp.configuration.announce import AnnounceIPv6
from exabgp.configuration.announce import AnnounceL2VPN
from exabgp.configuration.static import ParseStatic
from exabgp.configuration.static import ParseStaticRoute
from exabgp.configuration.flow import ParseFlow
from exabgp.configuration.flow import ParseFlowRoute
from exabgp.configuration.flow import ParseFlowThen
from exabgp.configuration.flow import ParseFlowMatch
from exabgp.configuration.flow import ParseFlowScope
from exabgp.configuration.l2vpn import ParseL2VPN
from exabgp.configuration.l2vpn import ParseVPLS
from exabgp.configuration.operational import ParseOperational

from exabgp.configuration.environment import environment

# for registration
from exabgp.configuration.announce.ip import AnnounceIP
from exabgp.configuration.announce.path import AnnouncePath
from exabgp.configuration.announce.label import AnnounceLabel
from exabgp.configuration.announce.vpn import AnnounceVPN
from exabgp.configuration.announce.flow import AnnounceFlow
from exabgp.configuration.announce.vpls import AnnounceVPLS


if sys.version_info[0] >= 3:
	StandardError = Exception


class _Configuration (object):
	def __init__ (self):
		self.processes = {}
		self.neighbors = {}
		self.logger = Logger ()

	def inject_change (self, peers, change):
		result = True
		for neighbor in self.neighbors:
			if neighbor in peers:
				if change.nlri.family() in self.neighbors[neighbor].families():
					self.neighbors[neighbor].rib.outgoing.add_to_rib(change)
				else:
					self.logger.error('the route family is not configured on neighbor','configuration')
					result = False
		return result

	def inject_eor (self, peers, family):
		result = False
		for neighbor in self.neighbors:
			if neighbor in peers:
				result = True
				self.neighbors[neighbor].eor.append(family)
		return result

	def inject_operational (self, peers, operational):
		result = True
		for neighbor in self.neighbors:
			if neighbor in peers:
				if operational.family() in self.neighbors[neighbor].families():
					if operational.name == 'ASM':
						self.neighbors[neighbor].asm[operational.family()] = operational
					self.neighbors[neighbor].messages.append(operational)
				else:
					self.logger.error('the route family is not configured on neighbor','configuration')
					result = False
		return result

	def inject_refresh (self, peers, refresh):
		result = True
		for neighbor in self.neighbors:
			if neighbor in peers:
				family = (refresh.afi,refresh.safi)
				if family in self.neighbors[neighbor].families():
					self.neighbors[neighbor].refresh.append(refresh.__class__(refresh.afi,refresh.safi))
				else:
					result = False
		return result


class Configuration (_Configuration):
	def __init__ (self, configurations, text=False):
		_Configuration.__init__(self)
		self.api_encoder = environment.settings().api.encoder

		self._configurations = configurations
		self._text = text

		self.error  = Error  ()
		self.scope  = Scope  ()

		self.tokeniser = Tokeniser(self.scope,self.error,self.logger)

		params = (self.tokeniser,self.scope,self.error,self.logger)
		generic                  = Section               (*params)
		self.process             = ParseProcess          (*params)
		self.template            = ParseTemplate         (*params)
		self.template_neighbor   = ParseTemplateNeighbor (*params)
		self.neighbor            = ParseNeighbor         (*params)
		self.family              = ParseFamily           (*params)
		self.addpath             = ParseAddPath          (*params)
		self.capability          = ParseCapability       (*params)
		self.api                 = ParseAPI              (*params)
		self.api_send            = ParseSend             (*params)
		self.api_receive         = ParseReceive          (*params)
		self.static              = ParseStatic           (*params)
		self.static_route        = ParseStaticRoute      (*params)
		self.announce            = SectionAnnounce       (*params)
		self.announce_ipv4       = AnnounceIPv4          (*params)
		self.announce_ipv6       = AnnounceIPv6          (*params)
		self.announce_l2vpn      = AnnounceL2VPN         (*params)
		self.flow                = ParseFlow             (*params)
		self.flow_route          = ParseFlowRoute        (*params)
		self.flow_match          = ParseFlowMatch        (*params)
		self.flow_then           = ParseFlowThen         (*params)
		self.flow_scope          = ParseFlowScope        (*params)
		self.l2vpn               = ParseL2VPN            (*params)
		self.vpls                = ParseVPLS             (*params)
		self.operational         = ParseOperational      (*params)

		# We should check if name are unique when running Section.__init__

		self._structure = {
			'root': {
				'class':    generic,
				'commands': [],
				'sections': {
					'process': self.process.name,
					'neighbor': self.neighbor.name,
					'template': self.template.name,
				},
			},
			self.process.name: {
				'class':    self.process,
				'commands': self.process.known.keys(),
				'sections': {},
			},
			self.template.name: {
				'class':    self.template,
				'commands': self.template.known.keys(),
				'sections': {
					'neighbor':    self.template_neighbor.name,
				},
			},
			self.template_neighbor.name: {
				'class':    self.template_neighbor,
				'commands': self.template_neighbor.known.keys(),
				'sections': {
					'family':      self.family.name,
					'capability':  self.capability.name,
					'add-path':    self.addpath.name,
					'api':         self.api.name,
					'static':      self.static.name,
					'flow':        self.flow.name,
					'l2vpn':       self.l2vpn.name,
					'operational': self.operational.name,
					'announce':    self.announce.name,
				},
			},
			self.neighbor.name: {
				'class':    self.neighbor,
				'commands': self.neighbor.known.keys(),
				'sections': {
					'family':      self.family.name,
					'capability':  self.capability.name,
					'add-path':    self.addpath.name,
					'api':         self.api.name,
					'static':      self.static.name,
					'flow':        self.flow.name,
					'l2vpn':       self.l2vpn.name,
					'operational': self.operational.name,
					'announce':    self.announce.name,
				},
			},
			self.family.name: {
				'class':    self.family,
				'commands': self.family.known.keys(),
				'sections': {
				},
			},
			self.capability.name: {
				'class':    self.capability,
				'commands': self.capability.known.keys(),
				'sections': {
				},
			},
			self.addpath.name: {
				'class':    self.addpath,
				'commands': self.addpath.known.keys(),
				'sections': {
				},
			},
			self.api.name: {
				'class':    self.api,
				'commands': self.api.known.keys(),
				'sections': {
					'send':    self.api_send.name,
					'receive': self.api_receive.name,
				},
			},
			self.api_send.name: {
				'class':    self.api_send,
				'commands': self.api_send.known.keys(),
				'sections': {
				},
			},
			self.api_receive.name: {
				'class':    self.api_receive,
				'commands': self.api_receive.known.keys(),
				'sections': {
				},
			},
			self.announce.name: {
				'class':    self.announce,
				'commands': self.announce.known.keys(),
				'sections': {
					'ipv4': self.announce_ipv4.name,
					'ipv6': self.announce_ipv6.name,
					'l2vpn': self.announce_l2vpn.name,
				},
			},
			self.announce_ipv4.name: {
				'class':    self.announce_ipv4,
				'commands': ['unicast', 'multicast', 'nlri-mpls', 'mpls-vpn', 'flow', 'flow-vpn'],
				'sections': {
				},
			},
			self.announce_ipv6.name: {
				'class':    self.announce_ipv6,
				'commands': ['unicast', 'multicast', 'nlri-mpls', 'mpls-vpn', 'flow', 'flow-vpn'],
				'sections': {
				},
			},
			self.announce_l2vpn.name: {
				'class':    self.announce_l2vpn,
				'commands': ['vpls',],
				'sections': {
				},
			},
			self.static.name: {
				'class':    self.static,
				'commands': ['route','attributes'],
				'sections': {
					'route': self.static_route.name,
				},
			},
			self.static_route.name: {
				'class':    self.static_route,
				'commands': self.static_route.known.keys(),
				'sections': {
				},
			},
			self.flow.name: {
				'class':    self.flow,
				'commands': self.flow.known.keys(),
				'sections': {
					'route': self.flow_route.name,
				},
			},
			self.flow_route.name: {
				'class':    self.flow_route,
				'commands': self.flow_route.known.keys(),
				'sections': {
					'match': self.flow_match.name,
					'then':  self.flow_then.name,
					'scope': self.flow_scope.name,
				},
			},
			self.flow_match.name: {
				'class':    self.flow_match,
				'commands': self.flow_match.known.keys(),
				'sections': {
				},
			},
			self.flow_then.name: {
				'class':    self.flow_then,
				'commands': self.flow_then.known.keys(),
				'sections': {
				},
			},
			self.flow_scope.name: {
				'class':    self.flow_scope,
				'commands': self.flow_scope.known.keys(),
				'sections': {
				}
			},
			self.l2vpn.name: {
				'class':    self.l2vpn,
				'commands': self.l2vpn.known.keys(),
				'sections': {
					'vpls': self.vpls.name,
				},
			},
			self.vpls.name: {
				'class':    self.vpls,
				'commands': self.l2vpn.known.keys(),
				'sections': {
				},
			},
			self.operational.name: {
				'class':    self.operational,
				'commands': self.operational.known.keys(),
				'sections': {
				}
			},
		}

		self._neighbors = {}
		self._previous_neighbors = {}

	def _clear (self):
		self.processes = {}
		self._previous_neighbors = self.neighbors
		self.neighbors = {}
		self._neighbors = {}

	# clear the parser data (ie: free memory)
	def _cleanup (self):
		self.error.clear()
		self.tokeniser.clear()
		self.scope.clear()

		self.process.clear()
		self.template.clear()
		self.template_neighbor.clear()
		self.neighbor.clear()
		self.family.clear()
		self.capability.clear()
		self.api.clear()
		self.api_send.clear()
		self.api_receive.clear()
		self.announce_ipv6.clear()
		self.announce_ipv4.clear()
		self.announce_l2vpn.clear()
		self.announce.clear()
		self.static.clear()
		self.static_route.clear()
		self.flow.clear()
		self.flow_route.clear()
		self.flow_match.clear()
		self.flow_then.clear()
		self.flow_scope.clear()
		self.l2vpn.clear()
		self.vpls.clear()
		self.operational.clear()

	def _rollback_reload (self):
		self.neighbors = self._previous_neighbors
		self.processes = self.process.processes
		self._neighbors = {}
		self._previous_neighbors = {}

	def _commit_reload (self):
		self.neighbors = self.neighbor.neighbors
		# XXX: Yes, we do not detect changes in processes and restart anything ..
		# XXX: This is a bug ..
		self.processes = self.process.processes
		self._neighbors = {}

		# Add the changes prior to the reload to the neighbor to correct handling of deleted routes
		for neighbor in self.neighbors:
			if neighbor in self._previous_neighbors:
				self.neighbors[neighbor].backup_changes = self._previous_neighbors[neighbor].changes

		self._previous_neighbors = {}
		self._cleanup()

	def reload (self):
		try:
			return self._reload()
		except KeyboardInterrupt:
			return self.error.set('configuration reload aborted by ^C or SIGINT')
		except Error as exc:
			if environment.settings().debug.configuration:
				raise
			return self.error.set(
				'problem parsing configuration file line %d\n'
				'error message: %s' % (self.tokeniser.index_line, exc)
			)
		except StandardError as exc:
			if environment.settings().debug.configuration:
				raise
			return self.error.set(
				'problem parsing configuration file line %d\n'
				'error message: %s' % (self.tokeniser.index_line, exc)
			)

	def _reload (self):
		# taking the first configuration available (FIFO buffer)
		fname = self._configurations.pop(0)
		self._configurations.append(fname)

		# clearing the current configuration to be able to re-parse it
		self._clear()

		if self._text:
			if not self.tokeniser.set_text(fname):
				return False
		else:
			if not self.tokeniser.set_file(fname):
				return False

		if self.section('root') is not True:
			# XXX: Should it be in neighbor ?
			self.process.add_api()
			self._rollback_reload()

			return self.error.set(
				"\n"
				"syntax error in section %s\n"
				"line %d: %s\n"
				"\n%s" % (
					self.scope.location(),
					self.tokeniser.number,
					' '.join(self.tokeniser.line),
					str(self.error)
				)
			)

		self.process.add_api()
		self._commit_reload()
		self._link()
		self.debug_check_route()
		self.debug_self_check()
		return True

	def _link (self):
		for neighbor in six.itervalues(self.neighbors):
			api = neighbor.api
			for process in api.get('processes',[]):
				self.processes.setdefault(process,{})['neighbor-changes'] = api['neighbor-changes']
				self.processes.setdefault(process,{})['negotiated'] = api['negotiated']
				self.processes.setdefault(process,{})['fsm'] = api['fsm']
				self.processes.setdefault(process,{})['signal'] = api['signal']
				for way in ('send','receive'):
					for name in ('parsed','packets','consolidate'):
						key = "%s-%s" % (way,name)
						if api[key]:
							self.processes[process].setdefault(key,[]).append(neighbor.router_id)
					for name in ('open', 'update', 'notification', 'keepalive', 'refresh', 'operational'):
						key = "%s-%s" % (way,name)
						if api[key]:
							self.processes[process].setdefault(key,[]).append(neighbor.router_id)

	def partial (self, section, text):
		self._cleanup()  # this perform a big cleanup (may be able to be smarter)
		self._clear()
		self.tokeniser.set_api(text if text.endswith(';') or text.endswith('}') else text + ' ;')

		if self.section(section) is not True:
			self._rollback_reload()
			self.logger.debug(
				"\n"
				"syntax error in api command %s\n"
				"line %d: %s\n"
				"\n%s" % (
					self.scope.location(),
					self.tokeniser.number,
					' '.join(self.tokeniser.line),
					str(self.error)
				),
				'configuration'
			)
			return False
		return True

	def _enter (self,name):
		location = self.tokeniser.iterate()
		self.logger.debug("> %-16s | %s" % (location,self.tokeniser.params()),'configuration')

		if location not in self._structure[name]['sections']:
			return self.error.set('section %s is invalid in %s, %s' % (location,name,self.scope.location()))

		self.scope.enter(location)
		self.scope.to_context()

		class_name = self._structure[name]['sections'][location]
		instance = self._structure[class_name].get('class',None)
		if not instance:
			raise RuntimeError('This should not be happening, debug time !')

		if not instance.pre():
			return False

		if not self.dispatch(self._structure[name]['sections'][location]):
			return False

		if not instance.post():
			return False

		left = self.scope.leave()
		if not left:
			return self.error.set('closing too many parenthesis')
		self.scope.to_context()

		self.logger.debug("< %-16s | %s" % (left,self.tokeniser.params()),'configuration')
		return True

	def _run (self,name):
		command = self.tokeniser.iterate()
		self.logger.debug(". %-16s | %s" % (command,self.tokeniser.params()),'configuration')

		if not self.run(name,command):
			return False
		return True

	def dispatch (self,name):
		while True:
			self.tokeniser()

			if self.tokeniser.end == ';':
				if self._run(name):
					continue
				return False

			if self.tokeniser.end == '{':
				if self._enter(name):
					continue
				return False

			if self.tokeniser.end == '}':
				return True

			if not self.tokeniser.end:  # finished
				return True

			return self.error.set('invalid syntax line %d' % self.tokeniser.index_line)
		return False

	def section (self, name):
		if name not in self._structure:
			return self.error.set('option %s is not allowed here' % name)

		return self.dispatch(name)

	def run (self, name, command):
		# restore 'anounce attribute' to provide backward 3.4 compatibility
		if name == 'static' and command == 'attribute':
			command = 'attributes'
		if command not in self._structure[name]['commands']:
			return self.error.set('invalid keyword "%s"' % command)

		return self._structure[name]['class'].parse(name,command)

	def debug_check_route (self):
		# we are not really running the program, just want to ....
		if environment.settings().debug.route:
			from exabgp.configuration.check import check_message
			if check_message(self.neighbors,environment.settings().debug.route):
				sys.exit(0)
			sys.exit(1)

	def debug_self_check (self):
		# we are not really running the program, just want check the configuration validity
		if environment.settings().debug.selfcheck:
			from exabgp.configuration.check import check_neighbor
			if check_neighbor(self.neighbors):
				sys.exit(0)
			sys.exit(1)



"""
multicast.py

Created by Thomas Morin on 2014-06-23.
Copyright (c) 2014-2017 Orange. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import IP
from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier import EthernetTag

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  IP Address Length (1 octet)          |
# +---------------------------------------+
# |   Originating Router's IP Addr        |
# |          (4 or 16 octets)             |
# +---------------------------------------+

# ===================================================================== EVPNNLRI


@EVPN.register
class Multicast (EVPN):
	CODE = 3
	NAME = "Inclusive Multicast Ethernet Tag"
	SHORT_NAME = "Multicast"

	def __init__ (self, rd, etag, ip, packed=None,nexthop=None,action=None,addpath=None):
		EVPN.__init__(self,action,addpath)
		self.nexthop = nexthop
		self.rd = rd
		self.etag = etag
		self.ip = ip
		self._pack(packed)

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return "%s:%s:%s:%s" % (
			self._prefix(),
			self.rd._str(),
			self.etag,
			self.ip,
		)

	def __hash__ (self):
		return hash((self.afi,self.safi,self.CODE,self.rd,self.etag,self.ip))

	def _pack (self, packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

		self._packed = concat_bytes(
			self.rd.pack(),
			self.etag.pack(),
			character(len(self.ip)*8),
			self.ip.pack()
		)
		return self._packed

	@classmethod
	def unpack (cls, data):
		rd = RouteDistinguisher.unpack(data[:8])
		etag = EthernetTag.unpack(data[8:12])
		iplen = ordinal(data[12])
		if iplen not in (4*8,16*8):
			raise Exception("IP len is %d, but EVPN route currently support only IPv4" % iplen)
		ip = IP.unpack(data[13:13+iplen//8])
		return cls(rd,etag,ip,data)

	def json (self, compact=None):
		content = ' "code": %d, ' % self.CODE
		content += '"parsed": true, '
		content += '"raw": "%s", ' % self._raw()
		content += '"name": "%s", ' % self.NAME
		content += '%s, ' % self.rd.json()
		content += self.etag.json()
		if self.ip:
			content += ', "ip": "%s"' % str(self.ip)
		return '{%s }' % content



#!/usr/bin/env python
# encoding: utf-8
"""
decode.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# pylint: ignore=E131
# pylama:ignore=E131

import sys

# INET include is required as it perform some AFI/SAFI registration
from exabgp.bgp.message.update.nlri import INET  # noqa

from exabgp.bgp.message.open.routerid import RouterID
from exabgp.protocol.ip import IPv4
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.holdtime import HoldTime

from exabgp.bgp.message import Update
from exabgp.bgp.message import Open
from exabgp.bgp.message.open import Version
from exabgp.bgp.message.open import ASN
from exabgp.bgp.message.open import RouterID
from exabgp.bgp.message.open import HoldTime
from exabgp.bgp.message.open.capability import Capabilities
from exabgp.bgp.message.open.capability import Capability
from exabgp.bgp.message.open.capability import Negotiated
from exabgp.bgp.message.update.nlri import NLRI


from exabgp.configuration.setup import environment
from exabgp.logger import Logger

from exabgp.util.test import data_from_body

environment.setup('')
logger = Logger()


bodies = []

body = [
	0x0, 0x0,  # len withdrawn routes
	# No routes to remove
	# Attributes
	0x0, 0x30,  # len attributes (48)
		0x40,  # Flag Transitive
		0x1,  # Code : Attribute ID Origin
		0x1,  # len
			0x0,  # Origin : IGP
		0x50,  # Flag Transitive + extended length
		0x2,  # Code : AS Path
		0x0, 0x16,  # len 22
			0x2,  # Type (AS_Sequence)
			0x5,  # length (in ASes as every asn is 0x0 0x0 prefixed ASN4 must have been negotiated)
				0x0, 0x0, 0xfe, 0xb0,  # ASN 65200
				0x0, 0x0, 0x78, 0x14,  # ASN 30740
				0x0, 0x0, 0x19, 0x35,  # ASN 6453
				0x0, 0x0, 0xb, 0x62,  # ASN 2914
				0x0, 0x0, 0x9, 0xd7,  # ASN 2519
		0x40,  # Flag Transitive
		0x3,  # Code: Next HOP
		0x4,  # len
			0x7f, 0x0, 0x0, 0x1,  # 127.0.0.1
		0xc0,  # 0x40 + 0x80 (Transitive Optional)
		0x08,  # Community
		0x8,  # Size 8
			0x78, 0x14, 0x19, 0x35,  # 30740:6453
			0x78, 0x14, 0xfd, 0xeb,  # 30740:65003
	# routes :
		0x18, 0x1, 0x0, 0x19,  # 1.0.25.0/24
		0x10, 0xde, 0xe6,  # 222.330.0.0/16
		0x11, 0xde, 0xe5, 0x80,
		0x12, 0xde, 0xe5, 0x0,
		0x10, 0xde, 0xe4,
		0x11, 0xdc, 0xf7, 0x0,
		0x11, 0xdc, 0x9e, 0x0,
		0x18, 0xdb, 0x79, 0xff,
		0x18, 0xdb, 0x79, 0xf9,
		0x16, 0xd8, 0xb3, 0xcc,
		0x18, 0xd8, 0xb3, 0xb6,
		0x17, 0xd8, 0xb3, 0xb4,
		0x16, 0xd8, 0xb3, 0xb0,
		0x18, 0xd8, 0xb3, 0x99,
		0x12, 0xd2, 0xaa, 0x0,
		0x11, 0xd2, 0x92, 0x80,
		0x11, 0xd2, 0x83, 0x80,
		0x13, 0xcb, 0x8c, 0x20,
		0x18, 0xca, 0xf5, 0xfe,
		0x18, 0xca, 0xf5, 0x8e,
		0x18, 0xca, 0xf3, 0xba,
		0x18, 0xca, 0xf0, 0x8d,
		0x12, 0xca, 0xef, 0xc0,
		0x12, 0xca, 0xe7, 0x40,
		0x10, 0xca, 0xd7,
		0x18, 0xca, 0xd2, 0x8,
		0x13, 0xca, 0xbd, 0xc0,
		0x14, 0xca, 0x58, 0x30,
		0x18, 0xca, 0x22, 0xbf,
		0x17, 0xca, 0x22, 0x96,
		0x18, 0xc0, 0x32, 0x6e,
		0x10, 0xb7, 0xb4,
		0x11, 0xb7, 0xb1, 0x80,
		0x10, 0xa3, 0x8b,
		0x11, 0x9d, 0x78, 0x80,
		0x15, 0x7c, 0xf1, 0x78,
		0x10, 0x7c, 0x6e,
		0x10, 0x7a, 0x67,
		0x10, 0x78, 0x33,
		0x10, 0x74, 0x5b,
		0x15, 0x73, 0xbb, 0x48,
		0x16, 0x73, 0xbb, 0x44,
		0x10, 0x73, 0xb3,
		0x11, 0x72, 0x45, 0x0,
		0x14, 0x71, 0x34, 0xf0,
		0x15, 0x70, 0x6d, 0x18,
		0x18, 0x67, 0xf6, 0xb3,
		0x17, 0x67, 0xb, 0x6,
		0x18, 0x67, 0x5, 0x75,
		0x18, 0x67, 0x5, 0x74,
		0x18, 0x67, 0x3, 0x10,
		0x13, 0x65, 0x37, 0xc0,
		0x11, 0x65, 0x32, 0x80,
		0x18, 0x65, 0x0, 0x1f,
		0x18, 0x65, 0x0, 0x1e,
		0x18, 0x65, 0x0, 0x1d,
		0x18, 0x65, 0x0, 0x1c,
		0xf, 0x24, 0x2,
		0x11, 0x1b, 0x79, 0x80,
		0x13, 0x1b, 0x60, 0x20,
		0x15, 0x1b, 0x60, 0x10,
		0x10, 0x1, 0x15,
		0x16, 0x1, 0x0, 0x1c,
		0x17, 0x1, 0x0, 0x1a
]

# asn4, data
bodies.append((True,body))

body = [
	0x0, 0x0,
	0x0, 0x38,
	0x40,
		0x1,
		0x1,
			0x0,
	0x50,
		0x2,
		0x0, 0x12,
			0x2,
			0x4,
				0x0, 0x0, 0xc7, 0x9d,
				0x0, 0x0, 0x9b, 0xbd,
				0x0, 0x0, 0xb, 0x62,
				0x0, 0x0, 0x9, 0xd7,
	0x40,
		0x3,
		0x4,
			0xb2, 0xd9, 0x76, 0x1,
	0xc0,
		0x8,
		0x14,
			0xb, 0x62, 0x1, 0x9a,
			0xb, 0x62, 0x5, 0x7b,
			0xb, 0x62, 0x9, 0x61,
			0xb, 0x62, 0xd, 0x48,
			0x9b, 0xbd, 0x1d, 0x1a,
	0x18, 0x1, 0x0, 0x19,
	0x10, 0xde, 0xe6,
	0x11, 0xde, 0xe5, 0x80,
	0x12, 0xde, 0xe5, 0x0,
	0x18, 0xdb, 0x79, 0xff,
	0x18, 0xdb, 0x79, 0xf9,
	0x16, 0xd8, 0xb3, 0xcc,
	0x18, 0xd8, 0xb3, 0xb6,
	0x17, 0xd8, 0xb3, 0xb4,
	0x16, 0xd8, 0xb3, 0xb0,
	0x18, 0xd8, 0xb3, 0x99,
	0x12, 0xd2, 0xaa, 0x0,
	0x11, 0xd2, 0x92, 0x80,
	0x11, 0xd2, 0x83, 0x80,
	0x13, 0xcb, 0x8c, 0x20,
	0x18, 0xca, 0xf5, 0xfe,
	0x18, 0xca, 0xf5, 0x8e,
	0x18, 0xca, 0xf3, 0xba,
	0x18, 0xca, 0xf0, 0x8d,
	0x12, 0xca, 0xef, 0xc0,
	0x12, 0xca, 0xe7, 0x40,
	0x18, 0xca, 0xd2, 0x8,
	0x13, 0xca, 0xbd, 0xc0,
	0x14, 0xca, 0x58, 0x30,
	0x18, 0xca, 0x22, 0xbf,
	0x17, 0xca, 0x22, 0x96,
	0x18, 0xc0, 0x32, 0x6e,
	0x10, 0xb7, 0xb4,
	0x11, 0x9d, 0x78, 0x80,
	0x15, 0x7c, 0xf1, 0x78,
	0x10, 0x7a, 0x67,
	0x15, 0x73, 0xbb, 0x48,
	0x16, 0x73, 0xbb, 0x44,
	0x11, 0x72, 0x45, 0x0,
	0x14, 0x71, 0x34, 0xf0,
	0x15, 0x70, 0x6d, 0x18,
	0x18, 0x67, 0xf6, 0xb3,
	0x17, 0x67, 0xb, 0x6,
	0x18, 0x67, 0x5, 0x75,
	0x18, 0x67, 0x5, 0x74,
	0x18, 0x67, 0x3, 0x10,
	0x13, 0x65, 0x37, 0xc0,
	0x11, 0x65, 0x32, 0x80,
	0x18, 0x65, 0x0, 0x1f,
	0x18, 0x65, 0x0, 0x1e,
	0x18, 0x65, 0x0, 0x1d,
	0x18, 0x65, 0x0, 0x1c,
	0xf, 0x24, 0x2,
	0x11, 0x1b, 0x79, 0x80,
	0x13, 0x1b, 0x60, 0x20,
	0x15, 0x1b, 0x60, 0x10,
	0x10, 0x1, 0x15,
	0x16, 0x1, 0x0, 0x1c,
	0x17, 0x1, 0x0, 0x1a
]
# asn4, data
bodies.append((True,body))

# BGP-LS with draft-gredler-idr-bgp-ls-segment-routing-ext
body = [0x00, 0x00, # withdrawn routes length 0
		0x00, 0xdc,	# Path Attribute Length 220
		0x90, 0x0e,
		0x00, 0x62,
		0x40, 0x04, 0x47, 0x04,
		0xac, 0x18, 0xf8, 0x4f,
		0x00, 0x00, 0x02, 0x00,
		0x55, 0x02, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x01, 0x00,
		0x00, 0x1a, 0x02, 0x00,
		0x00, 0x04, 0x00, 0x00,
		0xfd, 0xe8, 0x02, 0x01,
		0x00, 0x04, 0xac, 0x10,
		0x00, 0x58, 0x02, 0x03,
		0x00, 0x06, 0x17, 0x20,
		0x00, 0x16, 0x00, 0x02,
		0x01, 0x01, 0x00, 0x1a,
		0x02, 0x00, 0x00, 0x04,
		0x00, 0x00, 0xfd, 0xe8,
		0x02, 0x01, 0x00, 0x04,
		0xac, 0x10, 0x00, 0x58,
		0x02, 0x03, 0x00, 0x06,
		0x17, 0x20, 0x00, 0x16,
		0x00, 0x01, 0x01, 0x03,
		0x00, 0x04, 0x0a, 0x00,
		0x00, 0x07, 0x01, 0x04,
		0x00, 0x04, 0x0a, 0x00,
		0x00, 0x06, 0x40, 0x01,
		0x01, 0x00, 0x40, 0x02,
		0x06, 0x02, 0x01, 0x00,
		0x00, 0xfd, 0xe8,
		0x80, 0x1d, 			# Link State Attribute Code 29
		0x66,  					# Length
		0x04, 0x04, 0x00, 0x04, # Type 1028, Length 4
		0xac, 0x10, 0x00, 0x02, # IPv4 Router-ID 172.16.0.2
		0x04, 0x06,   			# Type 1030
		0x00, 0x04,
		0xac, 0x10, 0x00, 0x01, # IPv4 Router ID 172.16.0.1
		0x04, 0x40,				# Type 1088
		0x00, 0x04,
		0x00, 0x00, 0x00, 0x00, # SRLG = 0
		0x04, 0x41,				# Type 1089
		0x00, 0x04,
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Maximum Link bandwidth
		0x04, 0x42,				# Type 1090
		0x00, 0x04,
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Maximum reservable link bandwidth
		0x04, 0x43,				# Type 1091
		0x00, 0x20,
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P0
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P1
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P2
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P3
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P4
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P5
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P6
		0x4c, 0xee, 0x6b, 0x28, # 1000.00 Mbps Unreserved Bandwidth P7
		0x04, 0x44,				# Type 1092
		0x00, 0x04,				# Length 4
		0x00, 0x00, 0x00, 0x0a, # 10
		0x04, 0x47,				# Type 1095 Metric TLV
		0x00, 0x03,				# Length 3
		0x00, 0x00, 0x0a,		# 10
		0x04, 0x4b,				# Type 1099 Adjacency Segment TLV
		0x00, 0x07,				# Length 7
		0x30,					# Flags  0011 0000
		0x00,					# Weigth 0
		0x00, 0x00,				# Reserved 0x0
		0x04, 0x93, 0x1e]		# SID/Label 299806


# asn4, data
bodies.append((True,body))


class FakeNeighbor (object):
	description = 'a test neighbor'
	router_id = RouterID('127.0.0.1')
	local_address = IPv4('127.0.0.1')
	peer_address = IPv4('127.0.0.1')
	host_name = 'localhost'
	domain_name = 'localdomain'
	peer_as = ASN('65500')
	local_as = ASN('65500')
	hold_time = HoldTime(180)
	asn4 = False
	add_path = 0
	extended_message = False

	# capability
	route_refresh = False
	graceful_restart = False
	multisession = None
	add_path = None
	aigp = None
	operational = None

	@staticmethod
	def families ():
		return NLRI.known_families()

import unittest

# from contextlib import contextmanager
# @contextmanager
# def change_cwd (directory):
#     current_dir = os.path.abspath(os.path.dirname(__file__))
#     new_dir = os.path.join(current_dir, directory)
#     old_cwd = os.getcwd()
#     os.chdir(new_dir)
#     try:
#         yield
#     finally:
#         os.chdir(old_cwd)


class TestUpdateDecoding (unittest.TestCase):
	def setUp (self):
		self.negotiated = {}

		for asn4 in (True,False):
			neighbor = FakeNeighbor()
			neighbor.asn4 = asn4

			capa = Capabilities().new(neighbor,False)
			capa[Capability.CODE.MULTIPROTOCOL] = neighbor.families()

			# path = {}
			# for f in NLRI.known_families():
			# 	if neighbor.add_path:
			# 		path[f] = neighbor.add_path
			# capa[Capability.CODE.ADD_PATH] = path

			routerid_1 = str(neighbor.router_id)
			routerid_2 = '.'.join(str((int(_)+1) % 250) for _ in str(neighbor.router_id).split('.',-1))

			o1 = Open(Version(4),ASN(neighbor.local_as),HoldTime(180),RouterID(routerid_1),capa)
			o2 = Open(Version(4),ASN(neighbor.peer_as),HoldTime(180),RouterID(routerid_2),capa)

			negotiated = Negotiated(neighbor)
			negotiated.sent(o1)
			negotiated.received(o2)

			self.negotiated[asn4] = negotiated

	# size = len(body)+19
	# header = [0xFF,]*16 + [size >> 8] + [size & 0xFF] + [0x02]
	# print ''.join('%02X' % _ for _ in header+body

	def test_decoding_udpate_asn (self):
		for asn4,body in bodies:
			if asn4:
				continue
			Update.unpack_message(data_from_body(body),self.negotiated[asn4])

	def test_decoding_udpate_asn4 (self):
		for asn4,body in bodies:
			if not asn4:
				continue
			Update.unpack_message(data_from_body(body),self.negotiated[asn4])

if __name__ == '__main__':
	unittest.main()



# encoding: utf-8
"""
direction.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# =================================================================== Direction
#


class OUT (object):
	UNSET    = 0x00
	ANNOUNCE = 0x01
	WITHDRAW = 0x02


class IN (object):
	UNSET     = 0x00
	ANNOUNCED = 0x01
	WITHDRAWN = 0x02
